# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import torch
import torch.nn as nn
from mmcls.models.builder import HEADS, build_loss
from mmcls.models.heads import MultiLabelClsHead
from mmcv.cnn import normal_init


@HEADS.register_module()
class CustomHierarchicalLinearClsHead(MultiLabelClsHead):
    """Custom Linear classification head for hierarchical classification task.
    Args:
        num_classes (int): Number of categories.
        in_channels (int): Number of channels in the input feature map.
        loss (dict): Config of classification loss.
        multilabel_loss (dict): Config of multi-label classification loss.
    """

    def __init__(
        self,
        num_classes,
        in_channels,
        loss=dict(type="CrossEntropyLoss", use_sigmoid=True, reduction="mean", loss_weight=1.0),
        multilabel_loss=dict(type="AsymmetricLoss", reduction="mean", loss_weight=1.0),
        **kwargs,
    ):
        self.hierarchical_info = kwargs.pop("hierarchical_info", None)
        assert self.hierarchical_info
        super(CustomHierarchicalLinearClsHead, self).__init__(loss=loss)
        if self.hierarchical_info["num_multiclass_heads"] + self.hierarchical_info["num_multilabel_classes"] == 0:
            raise ValueError("Invalid classification heads configuration")
        self.compute_multilabel_loss = False
        if self.hierarchical_info["num_multilabel_classes"] > 0:
            self.compute_multilabel_loss = build_loss(multilabel_loss)

        if num_classes <= 0:
            raise ValueError(f"num_classes={num_classes} must be a positive integer")

        self.in_channels = in_channels
        self.num_classes = num_classes
        self._init_layers()

    def _init_layers(self):
        self.fc = nn.Linear(self.in_channels, self.num_classes)

    def init_weights(self):
        normal_init(self.fc, mean=0, std=0.01, bias=0)

    def loss(self, cls_score, gt_label, multilabel=False, valid_label_mask=None):
        num_samples = len(cls_score)
        # compute loss
        if multilabel:
            gt_label = gt_label.type_as(cls_score)
            # map difficult examples to positive ones
            _gt_label = torch.abs(gt_label)

            loss = self.compute_multilabel_loss(
                cls_score, _gt_label, valid_label_mask=valid_label_mask, avg_factor=num_samples
            )
        else:
            loss = self.compute_loss(cls_score, gt_label, avg_factor=num_samples)

        return loss

    def forward_train(self, x, gt_label, **kwargs):
        img_metas = kwargs.get("img_metas", None)
        gt_label = gt_label.type_as(x)
        cls_score = self.fc(x)

        losses = dict(loss=0.0)
        for i in range(self.hierarchical_info["num_multiclass_heads"]):
            head_gt = gt_label[:, i]
            head_logits = cls_score[
                :,
                self.hierarchical_info["head_idx_to_logits_range"][i][0] : self.hierarchical_info[
                    "head_idx_to_logits_range"
                ][i][1],
            ]
            valid_mask = head_gt >= 0
            head_gt = head_gt[valid_mask].long()
            head_logits = head_logits[valid_mask, :]
            multiclass_loss = self.loss(head_logits, head_gt)
            losses["loss"] += multiclass_loss

        if self.hierarchical_info["num_multiclass_heads"] > 1:
            losses["loss"] /= self.hierarchical_info["num_multiclass_heads"]

        if self.compute_multilabel_loss:
            head_gt = gt_label[:, self.hierarchical_info["num_multiclass_heads"] :]
            head_logits = cls_score[:, self.hierarchical_info["num_single_label_classes"] :]
            valid_batch_mask = head_gt >= 0
            head_gt = head_gt[
                valid_batch_mask,
            ]
            head_logits = head_logits[
                valid_batch_mask,
            ]

            # multilabel_loss is assumed to perform no batch averaging
            if img_metas is not None:
                valid_label_mask = self.get_valid_label_mask(img_metas=img_metas)[
                    :, self.hierarchical_info["num_single_label_classes"] :
                ]
                valid_label_mask = valid_label_mask.to(x.device)
                valid_label_mask = valid_label_mask[valid_batch_mask]
            else:
                valid_label_mask = None
            multilabel_loss = self.loss(head_logits, head_gt, multilabel=True, valid_label_mask=valid_label_mask)
            losses["loss"] += multilabel_loss.mean()
        return losses

    def simple_test(self, img):
        """Test without augmentation."""
        cls_score = self.fc(img)
        if isinstance(cls_score, list):
            cls_score = sum(cls_score) / float(len(cls_score))

        multiclass_logits = []
        for i in range(self.hierarchical_info["num_multiclass_heads"]):
            multiclass_logit = cls_score[
                :,
                self.hierarchical_info["head_idx_to_logits_range"][i][0] : self.hierarchical_info[
                    "head_idx_to_logits_range"
                ][i][1],
            ]
            if not torch.onnx.is_in_onnx_export():
                multiclass_logit = torch.softmax(multiclass_logit, dim=1)
            multiclass_logits.append(multiclass_logit)
        multiclass_pred = torch.cat(multiclass_logits, dim=1) if multiclass_logits else None

        if self.compute_multilabel_loss:
            multilabel_logits = cls_score[:, self.hierarchical_info["num_single_label_classes"] :]
            if not torch.onnx.is_in_onnx_export():
                multilabel_pred = torch.sigmoid(multilabel_logits) if multilabel_logits is not None else None
            else:
                multilabel_pred = multilabel_logits
            if multiclass_pred is not None:
                pred = torch.cat([multiclass_pred, multilabel_pred], axis=1)
            else:
                pred = multilabel_pred
        else:
            pred = multiclass_pred

        if torch.onnx.is_in_onnx_export():
            return pred
        pred = list(pred.detach().cpu().numpy())
        return pred

    def get_valid_label_mask(self, img_metas):
        valid_label_mask = []
        for i, meta in enumerate(img_metas):
            mask = torch.Tensor([1 for _ in range(self.num_classes)])
            if "ignored_labels" in meta and meta["ignored_labels"]:
                mask[meta["ignored_labels"]] = 0
            valid_label_mask.append(mask)
        valid_label_mask = torch.stack(valid_label_mask, dim=0)
        return valid_label_mask



_base_ = ["./data_seg.py"]

__dataset_type = "CustomDataset"
__data_root = "tests/assets/common_semantic_segmentation_dataset"

data = dict(
    train=dict(
        type="RepeatDataset",
        times=1,
        dataset=dict(
            type=__dataset_type,
            data_root=__data_root,
            img_dir="train/images",
            ann_dir="train/masks",
            classes=["background", "person"],
        ),
    ),
    val=dict(
        type=__dataset_type,
        data_root=__data_root,
        img_dir="val/images",
        ann_dir="val/masks",
        classes=["background", "person"],
    ),
    test=dict(
        type=__dataset_type,
        data_root=__data_root,
        img_dir="val/images",
        ann_dir="val/masks",
        classes=["background", "person"],
    ),
)



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from mmdet.models.builder import HEADS
from mmdet.models.dense_heads.retina_head import RetinaHead


@HEADS.register_module()
class CustomRetinaHead(RetinaHead):
    def __init__(self, *args, bg_loss_weight=-1.0, **kwargs):
        super().__init__(*args, **kwargs)
        self.bg_loss_weight = bg_loss_weight

    def loss_single(
        self, cls_score, bbox_pred, anchors, labels, label_weights, bbox_targets, bbox_weights, num_total_samples
    ):
        """Compute loss of a single scale level.

        Args:
            cls_score (Tensor): Box scores for each scale level
                Has shape (N, num_anchors * num_classes, H, W).
            bbox_pred (Tensor): Box energies / deltas for each scale
                level with shape (N, num_anchors * 4, H, W).
            anchors (Tensor): Box reference for each scale level with shape
                (N, num_total_anchors, 4).
            labels (Tensor): Labels of each anchors with shape
                (N, num_total_anchors).
            label_weights (Tensor): Label weights of each anchor with shape
                (N, num_total_anchors)
            bbox_targets (Tensor): BBox regression targets of each anchor wight
                shape (N, num_total_anchors, 4).
            bbox_weights (Tensor): BBox regression loss weights of each anchor
                with shape (N, num_total_anchors, 4).
            num_total_samples (int): If sampling, num total samples equal to
                the number of total anchors; Otherwise, it is the number of
                positive anchors.

        Returns:
            dict[str, Tensor]: A dictionary of loss components.
        """
        # classification loss
        labels = labels.reshape(-1)
        label_weights = label_weights.reshape(-1)
        cls_score = cls_score.permute(0, 2, 3, 1).reshape(-1, self.cls_out_channels)
        # Re-weigting BG loss
        if self.bg_loss_weight >= 0.0:
            neg_indices = labels == self.num_classes
            label_weights[neg_indices] = self.bg_loss_weight

        loss_cls = self.loss_cls(cls_score, labels, label_weights, avg_factor=num_total_samples)
        # regression loss
        bbox_targets = bbox_targets.reshape(-1, 4)
        bbox_weights = bbox_weights.reshape(-1, 4)
        bbox_pred = bbox_pred.permute(0, 2, 3, 1).reshape(-1, 4)
        if self.reg_decoded_bbox:
            # When the regression loss (e.g. `IouLoss`, `GIouLoss`)
            # is applied directly on the decoded bounding boxes, it
            # decodes the already encoded coordinates to absolute format.
            anchors = anchors.reshape(-1, 4)
            bbox_pred = self.bbox_coder.decode(anchors, bbox_pred)
        loss_bbox = self.loss_bbox(bbox_pred, bbox_targets, bbox_weights, avg_factor=num_total_samples)
        return loss_cls, loss_bbox



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from math import inf, isnan
from typing import Optional

from mmcv.runner import BaseRunner, LrUpdaterHook
from mmcv.runner.hooks import HOOKS, Hook
from mmcv.utils import print_log

from otx.mpa.utils.logger import get_logger

logger = get_logger()


# Temp copy from detection_task
# TODO: refactoing
class EarlyStoppingHook(Hook):
    """
    Cancel training when a metric has stopped improving.

    Early Stopping hook monitors a metric quantity and if no improvement is seen for a ‘patience’
    number of epochs, the training is cancelled.

    :param interval: the number of intervals for checking early stop. The interval number should be
                     the same as the evaluation interval - the `interval` variable set in
                     `evaluation` config.
    :param metric: the metric name to be monitored
    :param rule: greater or less.  In `less` mode, training will stop when the metric has stopped
                 decreasing and in `greater` mode it will stop when the metric has stopped
                 increasing.
    :param patience: Number of epochs with no improvement after which the training will be reduced.
                     For example, if patience = 2, then we will ignore the first 2 epochs with no
                     improvement, and will only cancel the training after the 3rd epoch if the
                     metric still hasn’t improved then
    :param iteration_patience: Number of iterations must be trained after the last improvement
                               before training stops. The same as patience but the training
                               continues if the number of iteration is lower than iteration_patience
                               This variable makes sure a model is trained enough for some
                               iterations after the last improvement before stopping.
    :param min_delta: Minimal decay applied to lr. If the difference between new and old lr is
                      smaller than eps, the update is ignored
    """

    rule_map = {"greater": lambda x, y: x > y, "less": lambda x, y: x < y}
    init_value_map = {"greater": -inf, "less": inf}
    greater_keys = ["acc", "top", "AR@", "auc", "precision", "mAP", "mDice", "mIoU", "mAcc", "aAcc", "MHAcc"]
    less_keys = ["loss"]

    def __init__(
        self,
        interval: int,
        metric: str = "bbox_mAP",
        rule: Optional[str] = None,
        patience: int = 5,
        iteration_patience: int = 500,
        min_delta: float = 0.0,
    ):
        super().__init__()
        self.patience = patience
        self.iteration_patience = iteration_patience
        self.interval = interval
        self.min_delta = min_delta
        self._init_rule(rule, metric)

        self.min_delta *= 1 if self.rule == "greater" else -1
        self.last_iter = 0
        self.wait_count = 0
        self.best_score = self.init_value_map[self.rule]

    def _init_rule(self, rule, key_indicator):
        """Initialize rule, key_indicator, comparison_func, and best score.

        Here is the rule to determine which rule is used for key indicator
        when the rule is not specific:
        1. If the key indicator is in ``self.greater_keys``, the rule will be
           specified as 'greater'.
        2. Or if the key indicator is in ``self.less_keys``, the rule will be
           specified as 'less'.
        3. Or if the key indicator is equal to the substring in any one item
           in ``self.greater_keys``, the rule will be specified as 'greater'.
        4. Or if the key indicator is equal to the substring in any one item
           in ``self.less_keys``, the rule will be specified as 'less'.

        Args:
            rule (str | None): Comparison rule for best score.
            key_indicator (str | None): Key indicator to determine the
                comparison rule.
        """
        if rule not in self.rule_map and rule is not None:
            raise KeyError(f"rule must be greater, less or None, " f"but got {rule}.")

        if rule is None:
            if key_indicator in self.greater_keys or any(key in key_indicator for key in self.greater_keys):
                rule = "greater"
            elif key_indicator in self.less_keys or any(key in key_indicator for key in self.less_keys):
                rule = "less"
            else:
                raise ValueError(
                    f"Cannot infer the rule for key " f"{key_indicator}, thus a specific rule " f"must be specified."
                )
        self.rule = rule
        self.key_indicator = key_indicator
        self.compare_func = self.rule_map[self.rule]

    def before_run(self, runner: BaseRunner):
        self.by_epoch = False if runner.max_epochs is None else True
        for hook in runner.hooks:
            if isinstance(hook, LrUpdaterHook):
                self.warmup_iters = hook.warmup_iters
                break
        if getattr(self, "warmup_iters", None) is None:
            raise ValueError("LrUpdaterHook must be registered to runner.")

    def after_train_iter(self, runner: BaseRunner):
        """Called after every training iter to evaluate the results."""
        if not self.by_epoch:
            self._do_check_stopping(runner)

    def after_train_epoch(self, runner: BaseRunner):
        """Called after every training epoch to evaluate the results."""
        if self.by_epoch:
            self._do_check_stopping(runner)

    def _do_check_stopping(self, runner):
        if not self._should_check_stopping(runner) or self.warmup_iters > runner.iter:
            return

        if runner.rank == 0:
            if self.key_indicator not in runner.log_buffer.output:
                raise KeyError(
                    f"metric {self.key_indicator} does not exist in buffer. Please check "
                    f"{self.key_indicator} is cached in evaluation output buffer"
                )

            key_score = runner.log_buffer.output[self.key_indicator]
            if self.compare_func(key_score - self.min_delta, self.best_score):
                self.best_score = key_score
                self.wait_count = 0
                self.last_iter = runner.iter
            else:
                self.wait_count += 1
                if self.wait_count >= self.patience:
                    if runner.iter - self.last_iter < self.iteration_patience:
                        print_log(
                            f"\nSkip early stopping. Accumulated iteration "
                            f"{runner.iter - self.last_iter} from the last "
                            f"improvement must be larger than {self.iteration_patience} to trigger "
                            f"Early Stopping.",
                            logger=runner.logger,
                        )
                        return
                    stop_point = runner.epoch if self.by_epoch else runner.iter
                    print_log(
                        f"\nEarly Stopping at :{stop_point} with " f"best {self.key_indicator}: {self.best_score}",
                        logger=runner.logger,
                    )
                    runner.should_stop = True

    def _should_check_stopping(self, runner):
        check_time = self.every_n_epochs if self.by_epoch else self.every_n_iters
        if not check_time(runner, self.interval):
            # No evaluation during the interval.
            return False
        return True


@HOOKS.register_module()
class LazyEarlyStoppingHook(EarlyStoppingHook):
    def __init__(
        self,
        interval: int,
        metric: str = "bbox_mAP",
        rule: str = None,
        patience: int = 5,
        iteration_patience: int = 500,
        min_delta: float = 0.0,
        start: int = None,
    ):
        self.start = start
        super(LazyEarlyStoppingHook, self).__init__(interval, metric, rule, patience, iteration_patience, min_delta)

    def _should_check_stopping(self, runner):
        if self.by_epoch:
            current = runner.epoch
            check_time = self.every_n_epochs
        else:
            current = runner.iter
            check_time = self.every_n_iters

        if self.start is None:
            if not check_time(runner, self.interval):
                # No evaluation during the interval.
                return False
        elif (current + 1) < self.start:
            return False
        else:
            if (current + 1 - self.start) % self.interval:
                return False
        return True


@HOOKS.register_module()
class ReduceLROnPlateauLrUpdaterHook(LrUpdaterHook):
    """
    Reduce learning rate when a metric has stopped improving.

    Models often benefit from reducing the learning rate by a factor of 2-10 once learning stagnates.
    This scheduler reads a metrics quantity and if no improvement is seen for a ‘patience’
    number of epochs, the learning rate is reduced.

    :param min_lr: minimum learning rate. The lower bound of the desired learning rate.
    :param interval: the number of intervals for checking the hook. The interval number should be
                     the same as the evaluation interval - the `interval` variable set in
                     `evaluation` config.
    :param metric: the metric name to be monitored
    :param rule: greater or less.  In `less` mode, learning rate will be dropped if the metric has
                 stopped decreasing and in `greater` mode it will be dropped when the metric has
                 stopped increasing.
    :param patience: Number of epochs with no improvement after which learning rate will be reduced.
                     For example, if patience = 2, then we will ignore the first 2 epochs with no
                     improvement, and will only drop LR after the 3rd epoch if the metric still
                     hasn’t improved then
    :param iteration_patience: Number of iterations must be trained after the last improvement
                               before LR drops. The same as patience but the LR remains the same if
                               the number of iteration is lower than iteration_patience. This
                               variable makes sure a model is trained enough for some iterations
                               after the last improvement before dropping the LR.
    :param factor: Factor to be multiply with the learning rate.
                   For example, new_lr = current_lr * factor
    """

    rule_map = {"greater": lambda x, y: x > y, "less": lambda x, y: x < y}
    init_value_map = {"greater": -inf, "less": inf}
    greater_keys = ["acc", "top", "AR@", "auc", "precision", "mAP", "mDice", "mIoU", "mAcc", "aAcc"]
    less_keys = ["loss"]

    def __init__(
        self,
        min_lr: float,
        interval: int,
        metric: str = "bbox_mAP",
        rule: Optional[str] = None,
        factor: float = 0.1,
        patience: int = 3,
        iteration_patience: int = 300,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.interval = interval
        self.min_lr = min_lr
        self.factor = factor
        self.patience = patience
        self.iteration_patience = iteration_patience
        self.metric = metric
        self.bad_count = 0
        self.last_iter = 0
        self.current_lr = None
        self._init_rule(rule, metric)
        self.best_score = self.init_value_map[self.rule]

    def _init_rule(self, rule, key_indicator):
        """Initialize rule, key_indicator, comparison_func, and best score.

        Here is the rule to determine which rule is used for key indicator
        when the rule is not specific:
        1. If the key indicator is in ``self.greater_keys``, the rule will be
           specified as 'greater'.
        2. Or if the key indicator is in ``self.less_keys``, the rule will be
           specified as 'less'.
        3. Or if the key indicator is equal to the substring in any one item
           in ``self.greater_keys``, the rule will be specified as 'greater'.
        4. Or if the key indicator is equal to the substring in any one item
           in ``self.less_keys``, the rule will be specified as 'less'.

        Args:
            rule (str | None): Comparison rule for best score.
            key_indicator (str | None): Key indicator to determine the
                comparison rule.
        """
        if rule not in self.rule_map and rule is not None:
            raise KeyError(f"rule must be greater, less or None, " f"but got {rule}.")

        if rule is None:
            if key_indicator in self.greater_keys or any(key in key_indicator for key in self.greater_keys):
                rule = "greater"
            elif key_indicator in self.less_keys or any(key in key_indicator for key in self.less_keys):
                rule = "less"
            else:
                raise ValueError(
                    f"Cannot infer the rule for key " f"{key_indicator}, thus a specific rule " f"must be specified."
                )
        self.rule = rule
        self.key_indicator = key_indicator
        self.compare_func = self.rule_map[self.rule]

    def _should_check_stopping(self, runner):
        check_time = self.every_n_epochs if self.by_epoch else self.every_n_iters
        if not check_time(runner, self.interval):
            # No evaluation during the interval.
            return False
        return True

    def get_lr(self, runner: BaseRunner, base_lr: float):
        if not self._should_check_stopping(runner) or self.warmup_iters > runner.iter:
            return self.current_lr if self.current_lr is not None else base_lr

        if self.current_lr is None:
            self.current_lr = base_lr

        if hasattr(runner, self.metric):
            score = getattr(runner, self.metric, 0.0)
        else:
            return self.current_lr

        print_log(
            f"\nBest Score: {self.best_score}, Current Score: {score}, Patience: {self.patience} "
            f"Count: {self.bad_count}",
            logger=runner.logger,
        )
        if self.compare_func(score, self.best_score):
            self.best_score = score
            self.bad_count = 0
            self.last_iter = runner.iter
        else:
            self.bad_count += 1

        if self.bad_count >= self.patience:
            if runner.iter - self.last_iter < self.iteration_patience:
                print_log(
                    f"\nSkip LR dropping. Accumulated iteration "
                    f"{runner.iter - self.last_iter} from the last "
                    f"improvement must be larger than {self.iteration_patience} to trigger "
                    f"LR dropping.",
                    logger=runner.logger,
                )
                return self.current_lr
            self.last_iter = runner.iter
            self.bad_count = 0
            print_log(
                f"\nDrop LR from: {self.current_lr}, to: " f"{max(self.current_lr * self.factor, self.min_lr)}",
                logger=runner.logger,
            )
            self.current_lr = max(self.current_lr * self.factor, self.min_lr)
        return self.current_lr

    def before_run(self, runner: BaseRunner):
        # TODO: remove overloaded method after fixing the issue
        #  https://github.com/open-mmlab/mmdetection/issues/6572
        for group in runner.optimizer.param_groups:
            group.setdefault("initial_lr", group["lr"])
        self.base_lr = [group["initial_lr"] for group in runner.optimizer.param_groups]
        self.bad_count = 0
        self.last_iter = 0
        self.current_lr = None
        self.best_score = self.init_value_map[self.rule]


@HOOKS.register_module()
class StopLossNanTrainingHook(Hook):
    def after_train_iter(self, runner: BaseRunner):
        if isnan(runner.outputs["loss"].item()):
            logger.warning("Early Stopping since loss is NaN")
            runner.should_stop = True



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import importlib
import json
import os
import os.path as osp
import re
from abc import ABC, abstractmethod
from collections import OrderedDict
from copy import deepcopy
from typing import List, Optional, Type

import pytest
import yaml

from otx.api.configuration.helper import create as otx_api_configuration_helper_create
from otx.api.entities.inference_parameters import InferenceParameters
from otx.api.entities.model import ModelEntity, ModelFormat, ModelOptimizationType
from otx.api.entities.model_template import parse_model_template
from otx.api.entities.optimization_parameters import OptimizationParameters
from otx.api.entities.resultset import ResultSetEntity
from otx.api.entities.subset import Subset
from otx.api.entities.task_environment import TaskEnvironment
from otx.api.serialization.label_mapper import LabelSchemaMapper, label_schema_to_bytes
from otx.api.usecases.adapters.model_adapter import ModelAdapter
from otx.api.usecases.tasks.interfaces.export_interface import ExportType
from otx.api.usecases.tasks.interfaces.optimization_interface import OptimizationType
from otx.api.utils.importing import get_impl_class

from .e2e_test_system import DataCollector
from .logging import get_logger
from .training_tests_common import (
    KEEP_CONFIG_FIELD_VALUE,
    performance_to_score_name_value,
)

logger = get_logger()


class BaseOTXTestAction(ABC):
    _name: Optional[str] = None
    _with_validation = False
    _depends_stages_names: List[str] = []

    def __init__(*args, **kwargs):
        pass

    @property
    def name(self):
        return type(self)._name

    @property
    def with_validation(self):
        return type(self)._with_validation

    @property
    def depends_stages_names(self):
        return type(self)._depends_stages_names

    def __str__(self):
        return (
            f"{type(self).__name__}("
            f"name={self.name}, "
            f"with_validation={self.with_validation}, "
            f"depends_stages_names={self.depends_stages_names})"
        )

    def _check_result_prev_stages(self, results_prev_stages, list_required_stages):
        for stage_name in list_required_stages:
            if not results_prev_stages or stage_name not in results_prev_stages:
                raise RuntimeError(
                    f"The action {self.name} requires results of the stage {stage_name}, " f"but they are absent"
                )

    @abstractmethod
    def __call__(self, data_collector: DataCollector, results_prev_stages: OrderedDict):
        raise NotImplementedError("The main action method is not implemented")


def create_environment_and_task(params, labels_schema, model_template, dataset=None, model_adapters=None):

    environment = TaskEnvironment(
        model=None,
        hyper_parameters=params,
        label_schema=labels_schema,
        model_template=model_template,
    )

    if model_adapters is not None:
        environment.model = ModelEntity(
            train_dataset=dataset,
            configuration=environment.get_model_configuration(),
            model_adapters=model_adapters,
        )

    logger.info("Create base Task")
    task_impl_path = model_template.entrypoints.base
    task_cls = get_impl_class(task_impl_path)
    task = task_cls(task_environment=environment)
    return environment, task


class OTXTestTrainingAction(BaseOTXTestAction):
    _name = "training"

    def __init__(
        self,
        dataset,
        labels_schema,
        template_path,
        num_training_iters,
        batch_size,
        checkpoint=None,
    ):
        self.dataset = dataset
        self.labels_schema = labels_schema
        self.template_path = template_path
        self.num_training_iters = num_training_iters
        self.batch_size = batch_size
        self.checkpoint = checkpoint

    def _get_training_performance_as_score_name_value(self):
        training_performance = getattr(self.output_model, "performance", None)
        if training_performance is None:
            raise RuntimeError("Cannot get training performance")
        return performance_to_score_name_value(training_performance)

    def _run_otx_training(self, data_collector):
        logger.debug(f"self.template_path = {self.template_path}")

        print(f"train dataset: {len(self.dataset.get_subset(Subset.TRAINING))} items")
        print(f"validation dataset: " f"{len(self.dataset.get_subset(Subset.VALIDATION))} items")

        logger.debug("Load model template")
        self.model_template = parse_model_template(self.template_path)

        logger.debug("Set hyperparameters")
        params = otx_api_configuration_helper_create(self.model_template.hyper_parameters.data)
        if self.num_training_iters != KEEP_CONFIG_FIELD_VALUE:
            params.learning_parameters.num_iters = int(self.num_training_iters)
            logger.debug(f"Set params.learning_parameters.num_iters=" f"{params.learning_parameters.num_iters}")
        else:
            logger.debug(f"Keep params.learning_parameters.num_iters=" f"{params.learning_parameters.num_iters}")

        if self.batch_size != KEEP_CONFIG_FIELD_VALUE:
            params.learning_parameters.batch_size = int(self.batch_size)
            logger.debug(f"Set params.learning_parameters.batch_size=" f"{params.learning_parameters.batch_size}")
        else:
            logger.debug(f"Keep params.learning_parameters.batch_size=" f"{params.learning_parameters.batch_size}")

        model_adapters = None
        if self.checkpoint is not None:
            logger.debug("Load pretrained model")
            model_adapters = {
                "weights.pth": ModelAdapter(open(self.checkpoint, "rb").read()),
            }
            label_schema_path = osp.join(osp.dirname(self.checkpoint), "label_schema.json")
            if osp.exists(label_schema_path):
                with open(label_schema_path, encoding="UTF-8") as read_file:
                    serialized_label_schema = LabelSchemaMapper.backward(json.load(read_file))
                model_adapters.update(
                    {"label_schema.json": ModelAdapter(label_schema_to_bytes(serialized_label_schema))}
                )

        self.environment, self.task = create_environment_and_task(
            params,
            self.labels_schema,
            self.model_template,
            self.dataset,
            model_adapters,
        )

        self.output_model = ModelEntity(
            self.dataset,
            self.environment.get_model_configuration(),
        )

        self.copy_hyperparams = deepcopy(self.task._hyperparams)

        logger.debug("Train model")

        try:
            self.task.train(self.dataset, self.output_model)
        except Exception as ex:
            raise RuntimeError("Training failed") from ex

        score_name, score_value = self._get_training_performance_as_score_name_value()
        logger.info(f"performance={self.output_model.performance}")
        data_collector.log_final_metric("metric_name", self.name + "/" + score_name)
        data_collector.log_final_metric("metric_value", score_value)

    def __call__(self, data_collector: DataCollector, results_prev_stages: OrderedDict):
        self._run_otx_training(data_collector)
        results = {
            "model_template": self.model_template,
            "task": self.task,
            "dataset": self.dataset,
            "environment": self.environment,
            "output_model": self.output_model,
        }
        return results


def is_nncf_enabled():
    return importlib.util.find_spec("nncf") is not None


def run_evaluation(dataset, task, model):
    logger.debug("Evaluation: Get predictions on the dataset")
    predicted_dataset = task.infer(dataset.with_empty_annotations(), InferenceParameters(is_evaluation=True))
    resultset = ResultSetEntity(
        model=model,
        ground_truth_dataset=dataset,
        prediction_dataset=predicted_dataset,
    )
    logger.debug("Evaluation: Estimate quality on dataset")
    task.evaluate(resultset)
    evaluation_performance = resultset.performance
    logger.info(f"Evaluation: performance={evaluation_performance}")
    score_name, score_value = performance_to_score_name_value(evaluation_performance)
    return score_name, score_value


class OTXTestTrainingEvaluationAction(BaseOTXTestAction):
    _name = "training_evaluation"
    _with_validation = True
    _depends_stages_names = ["training"]

    def __init__(self, subset=Subset.TESTING):
        self.subset = subset

    def _run_otx_evaluation(self, data_collector, dataset, task, trained_model):
        logger.info("Begin evaluation of trained model")
        validation_dataset = dataset.get_subset(self.subset)
        score_name, score_value = run_evaluation(validation_dataset, task, trained_model)
        data_collector.log_final_metric("metric_name", self.name + "/" + score_name)
        data_collector.log_final_metric("metric_value", score_value)
        logger.info(f"End evaluation of trained model, results: {score_name}: {score_value}")
        return score_name, score_value

    def __call__(self, data_collector: DataCollector, results_prev_stages: OrderedDict):
        self._check_result_prev_stages(results_prev_stages, self.depends_stages_names)

        kwargs = {
            "dataset": results_prev_stages["training"]["dataset"],
            "task": results_prev_stages["training"]["task"],
            "trained_model": results_prev_stages["training"]["output_model"],
        }

        score_name, score_value = self._run_otx_evaluation(data_collector, **kwargs)
        results = {"metrics": {"accuracy": {score_name: score_value}}}
        return results


def run_export(environment, dataset, task, action_name, expected_optimization_type):
    logger.debug(f'For action "{action_name}": Copy environment for evaluation exported model')

    environment_for_export = deepcopy(environment)

    logger.debug(f'For action "{action_name}": Create exported model')
    exported_model = ModelEntity(
        dataset,
        environment_for_export.get_model_configuration(),
    )
    logger.debug("Run export")

    try:
        task.export(ExportType.OPENVINO, exported_model)
    except Exception as ex:
        raise RuntimeError("Export to OpenVINO failed") from ex

    assert (
        exported_model.model_format == ModelFormat.OPENVINO
    ), f"In action '{action_name}': Wrong model format after export"
    assert (
        exported_model.optimization_type == expected_optimization_type
    ), f"In action '{action_name}': Wrong optimization type"

    logger.debug(f'For action "{action_name}": Set exported model into environment for export')
    environment_for_export.model = exported_model
    return environment_for_export, exported_model


class OTXTestExportAction(BaseOTXTestAction):
    _name = "export"
    _depends_stages_names = ["training"]

    def _run_otx_export(self, data_collector, environment, dataset, task):
        self.environment_for_export, self.exported_model = run_export(
            environment,
            dataset,
            task,
            action_name=self.name,
            expected_optimization_type=ModelOptimizationType.MO,
        )

    def __call__(self, data_collector: DataCollector, results_prev_stages: OrderedDict):
        self._check_result_prev_stages(results_prev_stages, self.depends_stages_names)

        kwargs = {
            "environment": results_prev_stages["training"]["environment"],
            "dataset": results_prev_stages["training"]["dataset"],
            "task": results_prev_stages["training"]["task"],
        }

        self._run_otx_export(data_collector, **kwargs)
        results = {
            "environment": self.environment_for_export,
            "exported_model": self.exported_model,
        }
        return results


def create_openvino_task(model_template, environment):
    logger.debug("Create OpenVINO Task")
    openvino_task_impl_path = model_template.entrypoints.openvino
    openvino_task_cls = get_impl_class(openvino_task_impl_path)
    openvino_task = openvino_task_cls(environment)
    return openvino_task


class OTXTestExportEvaluationAction(BaseOTXTestAction):
    _name = "export_evaluation"
    _with_validation = True
    _depends_stages_names = ["training", "export", "training_evaluation"]

    def __init__(self, subset=Subset.TESTING):
        self.subset = subset

    def _run_otx_export_evaluation(
        self,
        data_collector,
        model_template,
        dataset,
        environment_for_export,
        exported_model,
    ):
        logger.info("Begin evaluation of exported model")
        self.openvino_task = create_openvino_task(model_template, environment_for_export)
        validation_dataset = dataset.get_subset(self.subset)
        score_name, score_value = run_evaluation(validation_dataset, self.openvino_task, exported_model)
        data_collector.log_final_metric("metric_name", self.name + "/" + score_name)
        data_collector.log_final_metric("metric_value", score_value)
        logger.info("End evaluation of exported model")
        return score_name, score_value

    def __call__(self, data_collector: DataCollector, results_prev_stages: OrderedDict):
        self._check_result_prev_stages(results_prev_stages, self.depends_stages_names)

        kwargs = {
            "model_template": results_prev_stages["training"]["model_template"],
            "dataset": results_prev_stages["training"]["dataset"],
            "environment_for_export": results_prev_stages["export"]["environment"],
            "exported_model": results_prev_stages["export"]["exported_model"],
        }

        score_name, score_value = self._run_otx_export_evaluation(data_collector, **kwargs)
        results = {"metrics": {"accuracy": {score_name: score_value}}}
        return results


class OTXTestPotAction(BaseOTXTestAction):
    _name = "pot"
    _depends_stages_names = ["export"]

    def _run_otx_pot(self, data_collector, model_template, dataset, environment_for_export):
        logger.debug("Creating environment and task for POT optimization")
        self.environment_for_pot = deepcopy(environment_for_export)
        self.openvino_task_pot = create_openvino_task(model_template, environment_for_export)

        self.optimized_model_pot = ModelEntity(
            dataset,
            self.environment_for_pot.get_model_configuration(),
        )
        logger.info("Run POT optimization")

        try:
            self.openvino_task_pot.optimize(
                OptimizationType.POT,
                dataset,
                self.optimized_model_pot,
                OptimizationParameters(),
            )
        except Exception as ex:
            raise RuntimeError("POT optimization failed") from ex

        assert self.optimized_model_pot.model_format == ModelFormat.OPENVINO, "Wrong model format after pot"
        assert self.optimized_model_pot.optimization_type == ModelOptimizationType.POT, "Wrong optimization type"
        logger.info("POT optimization is finished")

    def __call__(self, data_collector: DataCollector, results_prev_stages: OrderedDict):
        self._check_result_prev_stages(results_prev_stages, self.depends_stages_names)

        kwargs = {
            "model_template": results_prev_stages["training"]["model_template"],
            "dataset": results_prev_stages["training"]["dataset"],
            "environment_for_export": results_prev_stages["export"]["environment"],
        }

        self._run_otx_pot(data_collector, **kwargs)
        results = {
            "openvino_task_pot": self.openvino_task_pot,
            "optimized_model_pot": self.optimized_model_pot,
        }
        return results


def check_fq_in_compressed_model(path_to_ref, compressed_type, model):
    """
    Check number of FakeQuantize nodes in the compressed model.
    """

    num_fq = len(re.findall(r'type="FakeQuantize"', model))

    assert os.path.exists(
        path_to_ref
    ), f"Reference file does not exist: {path_to_ref}. Current: {num_fq} for {compressed_type}."

    with open(path_to_ref, encoding="utf-8") as stream:
        ref_data = yaml.safe_load(stream)
    ref_num_fq = ref_data[compressed_type]["number_of_fakequantizers"]
    assert num_fq == ref_num_fq, f"Incorrect number of FQs in compressed model: {num_fq} != {ref_num_fq}"


class OTXTestPotValidationFQAction(BaseOTXTestAction):
    """
    Test to check number of FakeQuantize nodes in the compressed model by POT.
    """

    _name = "pot_validate_fq"
    _depends_stages_names = ["training", "pot"]

    def __init__(self, reference_dir):
        super().__init__()
        self.reference_dir = reference_dir

    def _run_otx_pot_validate_fq(self, optimized_model_pot):
        logger.info("Begin validation FQs of pot model")

        path_to_ref = os.path.join(self.reference_dir, "compressed_model.yml")

        check_fq_in_compressed_model(path_to_ref, "pot", optimized_model_pot.get_data("openvino.xml"))

        logger.info("End validation FQs of pot model")

    def __call__(self, data_collector: DataCollector, results_prev_stages: OrderedDict):
        self._check_result_prev_stages(results_prev_stages, self.depends_stages_names)

        kwargs = {
            "optimized_model_pot": results_prev_stages["pot"]["optimized_model_pot"],
        }

        self._run_otx_pot_validate_fq(**kwargs)

        return {}


class OTXTestPotEvaluationAction(BaseOTXTestAction):
    _name = "pot_evaluation"
    _with_validation = True
    _depends_stages_names = ["training", "pot", "export_evaluation"]

    def __init__(self, subset=Subset.TESTING):
        self.subset = subset

    def _run_otx_pot_evaluation(self, data_collector, dataset, openvino_task_pot, optimized_model_pot):
        logger.info("Begin evaluation of pot model")
        validation_dataset_pot = dataset.get_subset(self.subset)
        score_name, score_value = run_evaluation(validation_dataset_pot, openvino_task_pot, optimized_model_pot)
        data_collector.log_final_metric("metric_name", self.name + "/" + score_name)
        data_collector.log_final_metric("metric_value", score_value)
        logger.info("End evaluation of pot model")
        return score_name, score_value

    def __call__(self, data_collector: DataCollector, results_prev_stages: OrderedDict):
        self._check_result_prev_stages(results_prev_stages, self.depends_stages_names)

        kwargs = {
            "dataset": results_prev_stages["training"]["dataset"],
            "openvino_task_pot": results_prev_stages["pot"]["openvino_task_pot"],
            "optimized_model_pot": results_prev_stages["pot"]["optimized_model_pot"],
        }

        score_name, score_value = self._run_otx_pot_evaluation(data_collector, **kwargs)
        results = {"metrics": {"accuracy": {score_name: score_value}}}
        return results


class OTXTestNNCFAction(BaseOTXTestAction):
    _name = "nncf"
    _depends_stages_names = ["training"]

    def _run_otx_nncf(self, data_collector, model_template, dataset, trained_model, environment):
        logger.debug("Get predictions on the validation set for exported model")
        self.environment_for_nncf = deepcopy(environment)

        logger.info("Create NNCF Task")
        nncf_task_class_impl_path = model_template.entrypoints.nncf
        if not nncf_task_class_impl_path:
            pytest.skip("NNCF is not enabled for this template")

        if not is_nncf_enabled():
            pytest.skip("NNCF is not installed")

        logger.info("Creating NNCF task and structures")
        self.nncf_model = ModelEntity(
            dataset,
            self.environment_for_nncf.get_model_configuration(),
        )
        self.nncf_model.set_data("weights.pth", trained_model.get_data("weights.pth"))

        self.environment_for_nncf.model = self.nncf_model

        nncf_task_cls = get_impl_class(nncf_task_class_impl_path)
        self.nncf_task = nncf_task_cls(task_environment=self.environment_for_nncf)

        logger.info("Run NNCF optimization")
        try:
            self.nncf_task.optimize(OptimizationType.NNCF, dataset, self.nncf_model, None)
        except Exception as ex:
            raise RuntimeError("NNCF optimization failed") from ex

        assert self.nncf_model.optimization_type == ModelOptimizationType.NNCF, "Wrong optimization type"
        assert self.nncf_model.model_format == ModelFormat.BASE_FRAMEWORK, "Wrong model format"

        logger.info("NNCF optimization is finished")

    def __call__(self, data_collector: DataCollector, results_prev_stages: OrderedDict):
        self._check_result_prev_stages(results_prev_stages, self.depends_stages_names)

        kwargs = {
            "model_template": results_prev_stages["training"]["model_template"],
            "dataset": results_prev_stages["training"]["dataset"],
            "trained_model": results_prev_stages["training"]["output_model"],
            "environment": results_prev_stages["training"]["environment"],
        }

        self._run_otx_nncf(data_collector, **kwargs)
        results = {
            "nncf_task": self.nncf_task,
            "nncf_model": self.nncf_model,
            "nncf_environment": self.environment_for_nncf,
        }
        return results


# TODO: think about move to special file
def check_nncf_model_graph(model, path_to_dot):
    import networkx as nx

    logger.info(f"Reference graph: {path_to_dot}")
    load_graph = nx.drawing.nx_pydot.read_dot(path_to_dot)

    graph = model.get_graph()
    nx_graph = graph.get_graph_for_structure_analysis()

    for _, node in nx_graph.nodes(data=True):
        if "scope" in node:
            node.pop("scope")

    for k, attrs in nx_graph.nodes.items():
        attrs = {k: str(v) for k, v in attrs.items()}
        load_attrs = {k: str(v).strip('"') for k, v in load_graph.nodes[k].items()}
        if "scope" in load_attrs:
            load_attrs.pop("scope")
        if attrs != load_attrs:
            logger.info("ATTR: {} : {} != {}".format(k, attrs, load_attrs))
            return False

    return load_graph.nodes.keys() == nx_graph.nodes.keys() and nx.DiGraph(load_graph).edges == nx_graph.edges


class OTXTestNNCFGraphAction(BaseOTXTestAction):
    _name = "nncf_graph"

    def __init__(
        self,
        dataset,
        labels_schema,
        template_path,
        reference_dir,
        fn_get_compressed_model,
    ):
        self.dataset = dataset
        self.labels_schema = labels_schema
        self.template_path = template_path
        self.reference_dir = reference_dir
        self.fn_get_compressed_model = fn_get_compressed_model

    def _run_otx_nncf_graph(self, data_collector):
        # pylint:disable=protected-access
        logger.debug("Load model template")
        model_template = parse_model_template(self.template_path)
        nncf_task_class_impl_path = model_template.entrypoints.nncf

        if not nncf_task_class_impl_path:
            pytest.skip("NNCF is not enabled for this template")

        if not is_nncf_enabled():
            pytest.skip("NNCF is not installed")

        if not os.path.exists(self.reference_dir):
            pytest.skip("Reference directory does not exist")

        params = otx_api_configuration_helper_create(model_template.hyper_parameters.data)
        environment, task = create_environment_and_task(params, self.labels_schema, model_template)
        output_model = ModelEntity(
            self.dataset,
            environment.get_model_configuration(),
        )
        # Save model without training to create nncf_task
        task.save_model(output_model)

        logger.info("Create NNCF Task")
        environment_for_nncf = deepcopy(environment)

        logger.info("Creating NNCF task and structures")
        nncf_model = ModelEntity(
            self.dataset,
            environment_for_nncf.get_model_configuration(),
        )
        nncf_model.set_data("weights.pth", output_model.get_data("weights.pth"))

        environment_for_nncf.model = nncf_model

        nncf_task_cls = get_impl_class(nncf_task_class_impl_path)
        nncf_task = nncf_task_cls(task_environment=environment_for_nncf)

        path_to_ref_dot = os.path.join(self.reference_dir, "nncf", f"{nncf_task._nncf_preset}.dot")
        if not os.path.exists(path_to_ref_dot):
            pytest.skip("Reference file does not exist: {}".format(path_to_ref_dot))

        compressed_model = self.fn_get_compressed_model(nncf_task)

        assert check_nncf_model_graph(compressed_model, path_to_ref_dot), "Compressed model differs from the reference"

    def __call__(self, data_collector: DataCollector, results_prev_stages: OrderedDict):
        self._check_result_prev_stages(results_prev_stages, self.depends_stages_names)
        self._run_otx_nncf_graph(data_collector)
        return {}


class OTXTestNNCFEvaluationAction(BaseOTXTestAction):
    _name = "nncf_evaluation"
    _with_validation = True
    _depends_stages_names = ["training", "nncf", "training_evaluation"]

    def __init__(self, subset=Subset.TESTING):
        self.subset = subset

    def _run_otx_nncf_evaluation(self, data_collector, dataset, nncf_task, nncf_model):
        logger.info("Begin evaluation of nncf model")
        validation_dataset = dataset.get_subset(self.subset)
        score_name, score_value = run_evaluation(validation_dataset, nncf_task, nncf_model)
        data_collector.log_final_metric("metric_name", self.name + "/" + score_name)
        data_collector.log_final_metric("metric_value", score_value)
        logger.info("End evaluation of nncf model")
        return score_name, score_value

    def __call__(self, data_collector: DataCollector, results_prev_stages: OrderedDict):
        self._check_result_prev_stages(results_prev_stages, self.depends_stages_names)

        kwargs = {
            "dataset": results_prev_stages["training"]["dataset"],
            "nncf_task": results_prev_stages["nncf"]["nncf_task"],
            "nncf_model": results_prev_stages["nncf"]["nncf_model"],
        }

        score_name, score_value = self._run_otx_nncf_evaluation(data_collector, **kwargs)
        results = {"metrics": {"accuracy": {score_name: score_value}}}
        return results


class OTXTestNNCFExportAction(BaseOTXTestAction):
    _name = "nncf_export"
    _depends_stages_names = ["training", "nncf"]

    def __init__(self, subset=Subset.VALIDATION):
        self.subset = subset

    def _run_otx_nncf_export(self, data_collector, nncf_environment, dataset, nncf_task):
        logger.info("Begin export of nncf model")
        self.environment_nncf_export, self.nncf_exported_model = run_export(
            nncf_environment,
            dataset,
            nncf_task,
            action_name=self.name,
            expected_optimization_type=ModelOptimizationType.NNCF,
        )
        logger.info("End export of nncf model")

    def __call__(self, data_collector: DataCollector, results_prev_stages: OrderedDict):
        self._check_result_prev_stages(results_prev_stages, self.depends_stages_names)

        kwargs = {
            "nncf_environment": results_prev_stages["nncf"]["nncf_environment"],
            "dataset": results_prev_stages["training"]["dataset"],
            "nncf_task": results_prev_stages["nncf"]["nncf_task"],
        }

        self._run_otx_nncf_export(data_collector, **kwargs)
        results = {
            "environment": self.environment_nncf_export,
            "exported_model": self.nncf_exported_model,
        }
        return results


class OTXTestNNCFExportEvaluationAction(BaseOTXTestAction):
    _name = "nncf_export_evaluation"
    _with_validation = True
    _depends_stages_names = ["training", "nncf_export", "nncf_evaluation"]

    def __init__(self, subset=Subset.TESTING):
        self.subset = subset

    def _run_otx_nncf_export_evaluation(
        self,
        data_collector,
        model_template,
        dataset,
        nncf_environment_for_export,
        nncf_exported_model,
    ):
        logger.info("Begin evaluation of NNCF exported model")
        self.openvino_task = create_openvino_task(model_template, nncf_environment_for_export)
        validation_dataset = dataset.get_subset(self.subset)
        score_name, score_value = run_evaluation(validation_dataset, self.openvino_task, nncf_exported_model)
        data_collector.log_final_metric("metric_name", self.name + "/" + score_name)
        data_collector.log_final_metric("metric_value", score_value)
        logger.info("End evaluation of NNCF exported model")
        return score_name, score_value

    def __call__(self, data_collector: DataCollector, results_prev_stages: OrderedDict):
        self._check_result_prev_stages(results_prev_stages, self.depends_stages_names)

        kwargs = {
            "model_template": results_prev_stages["training"]["model_template"],
            "dataset": results_prev_stages["training"]["dataset"],
            "nncf_environment_for_export": results_prev_stages["nncf_export"]["environment"],
            "nncf_exported_model": results_prev_stages["nncf_export"]["exported_model"],
        }

        score_name, score_value = self._run_otx_nncf_export_evaluation(data_collector, **kwargs)
        results = {"metrics": {"accuracy": {score_name: score_value}}}
        return results


class OTXTestNNCFValidationFQAction(BaseOTXTestAction):
    """
    Test to check number of FakeQuantize nodes in the compressed model by POT.
    """

    _name = "nncf_validate_fq"
    _depends_stages_names = ["training", "nncf_export"]

    def __init__(self, reference_dir):
        super().__init__()
        self.reference_dir = reference_dir

    def _run_otx_nncf_validate_fq(self, nncf_exported_model):
        logger.info("Begin validation FQs of nncf model")

        path_to_ref = os.path.join(self.reference_dir, "compressed_model.yml")

        check_fq_in_compressed_model(path_to_ref, "nncf", nncf_exported_model.get_data("openvino.xml"))

        logger.info("End validation FQs of nncf model")

    def __call__(self, data_collector: DataCollector, results_prev_stages: OrderedDict):
        self._check_result_prev_stages(results_prev_stages, self.depends_stages_names)

        kwargs = {
            "nncf_exported_model": results_prev_stages["nncf_export"]["exported_model"],
        }

        self._run_otx_nncf_validate_fq(**kwargs)

        return {}


def get_default_test_action_classes() -> List[Type[BaseOTXTestAction]]:
    return [
        OTXTestTrainingAction,
        OTXTestTrainingEvaluationAction,
        OTXTestExportAction,
        OTXTestExportEvaluationAction,
        OTXTestPotAction,
        OTXTestPotEvaluationAction,
        OTXTestPotValidationFQAction,
        OTXTestNNCFAction,
        OTXTestNNCFEvaluationAction,
        OTXTestNNCFExportAction,
        OTXTestNNCFExportEvaluationAction,
        OTXTestNNCFGraphAction,
        OTXTestNNCFValidationFQAction,
    ]



"""Anomaly Classification / Detection / Segmentation Dataset Adapter."""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

# pylint: disable=invalid-name, too-many-locals, no-member
import os
from typing import Dict, List, Optional

import cv2
import numpy as np
from datumaro.components.dataset import Dataset as DatumaroDataset

from otx.algorithms.common.utils.mask_to_bbox import mask2bbox
from otx.api.entities.annotation import (
    Annotation,
    AnnotationSceneEntity,
    AnnotationSceneKind,
    NullAnnotationSceneEntity,
)
from otx.api.entities.dataset_item import DatasetItemEntity
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.id import ID
from otx.api.entities.image import Image
from otx.api.entities.label import LabelEntity
from otx.api.entities.scored_label import ScoredLabel
from otx.api.entities.shapes.rectangle import Rectangle
from otx.api.entities.subset import Subset
from otx.api.utils.segmentation_utils import create_annotation_from_segmentation_map
from otx.core.data.adapter.base_dataset_adapter import BaseDatasetAdapter


class AnomalyBaseDatasetAdapter(BaseDatasetAdapter):
    """BaseDataset Adpater for Anomaly tasks inherited from BaseDatasetAdapter."""

    def _import_dataset(
        self,
        train_data_roots: Optional[str] = None,
        val_data_roots: Optional[str] = None,
        test_data_roots: Optional[str] = None,
        unlabeled_data_roots: Optional[str] = None,
    ) -> Dict[Subset, DatumaroDataset]:
        """Import MVTec dataset.

        Args:
            train_data_roots (Optional[str]): Path for training data
            val_data_roots (Optional[str]): Path for validation data
            test_data_roots (Optional[str]): Path for test data
            unlabeled_data_roots (Optional[str]): Path for unlabeled data

        Returns:
            DatumaroDataset: Datumaro Dataset
        """
        # Construct dataset for training, validation, unlabeled
        # TODO: currently, only MVTec dataset format can be used
        dataset = {}
        if train_data_roots is None and test_data_roots is None:
            raise ValueError("At least 1 data_root is needed to train/test.")

        if train_data_roots:
            dataset[Subset.TRAINING] = DatumaroDataset.import_from(train_data_roots, format="image_dir")
            if val_data_roots:
                dataset[Subset.VALIDATION] = DatumaroDataset.import_from(val_data_roots, format="image_dir")
            else:
                raise NotImplementedError("Anomaly task needs validation dataset.")
        if test_data_roots:
            dataset[Subset.TESTING] = DatumaroDataset.import_from(test_data_roots, format="image_dir")
        return dataset

    def _prepare_anomaly_label_information(self) -> List[LabelEntity]:
        """Prepare LabelEntity List."""
        normal_label = LabelEntity(id=ID(0), name="Normal", domain=self.domain)
        abnormal_label = LabelEntity(
            id=ID(1),
            name="Anomalous",
            domain=self.domain,
            is_anomalous=True,
        )
        return [normal_label, abnormal_label]

    def get_otx_dataset(self) -> DatasetEntity:
        """Get DatasetEntity."""
        raise NotImplementedError()


class AnomalyClassificationDatasetAdapter(AnomalyBaseDatasetAdapter):
    """Anomaly classification adapter inherited from AnomalyBaseDatasetAdapter."""

    def get_otx_dataset(self) -> DatasetEntity:
        """Convert DatumaroDataset to DatasetEntity for Anomaly classification."""
        normal_label, abnormal_label = self._prepare_anomaly_label_information()
        self.label_entities = [normal_label, abnormal_label]

        # Prepare
        dataset_items: List[DatasetItemEntity] = []
        for subset, subset_data in self.dataset.items():
            for _, datumaro_items in subset_data.subsets().items():
                for datumaro_item in datumaro_items:
                    image = Image(file_path=datumaro_item.media.path)
                    label = normal_label if os.path.dirname(datumaro_item.id) == "good" else abnormal_label
                    shapes = [
                        Annotation(
                            Rectangle.generate_full_box(),
                            labels=[ScoredLabel(label=label, probability=1.0)],
                        )
                    ]
                    annotation_scene: Optional[AnnotationSceneEntity] = None
                    # Unlabeled dataset
                    if len(shapes) == 0:
                        annotation_scene = NullAnnotationSceneEntity()
                    else:
                        annotation_scene = AnnotationSceneEntity(
                            kind=AnnotationSceneKind.ANNOTATION, annotations=shapes
                        )
                    dataset_item = DatasetItemEntity(image, annotation_scene, subset=subset)
                    dataset_items.append(dataset_item)

        return DatasetEntity(items=dataset_items)


class AnomalyDetectionDatasetAdapter(AnomalyBaseDatasetAdapter):
    """Anomaly detection adapter inherited from AnomalyBaseDatasetAdapter."""

    def get_otx_dataset(self) -> DatasetEntity:
        """Conver DatumaroDataset to DatasetEntity for Anomaly detection."""
        normal_label, abnormal_label = self._prepare_anomaly_label_information()
        self.label_entities = [normal_label, abnormal_label]

        # Prepare
        dataset_items: List[DatasetItemEntity] = []
        for subset, subset_data in self.dataset.items():
            for _, datumaro_items in subset_data.subsets().items():
                for datumaro_item in datumaro_items:
                    image = Image(file_path=datumaro_item.media.path)
                    label = normal_label if os.path.dirname(datumaro_item.id) == "good" else abnormal_label
                    shapes = [
                        Annotation(
                            Rectangle.generate_full_box(),
                            labels=[ScoredLabel(label=label, probability=1.0)],
                        )
                    ]
                    # TODO: avoid hard coding, plan to enable MVTec to Datumaro
                    mask_file_path = os.path.join(
                        "/".join(datumaro_item.media.path.split("/")[:-3]),
                        "ground_truth",
                        str(datumaro_item.id) + "_mask.png",
                    )
                    if os.path.exists(mask_file_path):
                        mask = (cv2.imread(mask_file_path, cv2.IMREAD_GRAYSCALE) / 255).astype(np.uint8)
                        bboxes = mask2bbox(mask)
                        for bbox in bboxes:
                            x1, y1, x2, y2 = bbox
                            shapes.append(
                                Annotation(
                                    Rectangle(
                                        x1=x1 / image.width,
                                        y1=y1 / image.height,
                                        x2=x2 / image.width,
                                        y2=y2 / image.height,
                                    ),
                                    labels=[ScoredLabel(label=abnormal_label)],
                                )
                            )
                    annotation_scene: Optional[AnnotationSceneEntity] = None
                    # Unlabeled dataset
                    if len(shapes) == 0:
                        annotation_scene = NullAnnotationSceneEntity()
                    else:
                        annotation_scene = AnnotationSceneEntity(
                            kind=AnnotationSceneKind.ANNOTATION, annotations=shapes
                        )
                    dataset_item = DatasetItemEntity(image, annotation_scene, subset=subset)
                    dataset_items.append(dataset_item)

        return DatasetEntity(items=dataset_items)


class AnomalySegmentationDatasetAdapter(AnomalyBaseDatasetAdapter):
    """Anomaly segmentation adapter inherited by AnomalyBaseDatasetAdapter and BaseDatasetAdapter."""

    def get_otx_dataset(self) -> DatasetEntity:
        """Conver DatumaroDataset to DatasetEntity for Anomaly segmentation."""
        normal_label, abnormal_label = self._prepare_anomaly_label_information()
        self.label_entities = [normal_label, abnormal_label]

        # Prepare
        dataset_items: List[DatasetItemEntity] = []
        for subset, subset_data in self.dataset.items():
            for _, datumaro_items in subset_data.subsets().items():
                for datumaro_item in datumaro_items:
                    image = Image(file_path=datumaro_item.media.path)
                    label = normal_label if os.path.dirname(datumaro_item.id) == "good" else abnormal_label
                    shapes = [
                        Annotation(
                            Rectangle.generate_full_box(),
                            labels=[ScoredLabel(label=label, probability=1.0)],
                        )
                    ]
                    # TODO: avoid hard coding, plan to enable MVTec to Datumaro
                    mask_file_path = os.path.join(
                        "/".join(datumaro_item.media.path.split("/")[:-3]),
                        "ground_truth",
                        str(datumaro_item.id) + "_mask.png",
                    )
                    if os.path.exists(mask_file_path):
                        mask = (cv2.imread(mask_file_path, cv2.IMREAD_GRAYSCALE) / 255).astype(np.uint8)
                        shapes.extend(
                            create_annotation_from_segmentation_map(
                                hard_prediction=mask,
                                soft_prediction=np.ones_like(mask),
                                label_map={0: normal_label, 1: abnormal_label},
                            )
                        )
                    annotation_scene: Optional[AnnotationSceneEntity] = None
                    # Unlabeled dataset
                    if len(shapes) == 0:
                        annotation_scene = NullAnnotationSceneEntity()
                    else:
                        annotation_scene = AnnotationSceneEntity(
                            kind=AnnotationSceneKind.ANNOTATION, annotations=shapes
                        )
                    dataset_item = DatasetItemEntity(image, annotation_scene, subset=subset)
                    dataset_items.append(dataset_item)

        return DatasetEntity(items=dataset_items)



"""Time monitor callback module."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

# pylint: disable=too-many-instance-attributes,too-many-arguments

import logging
import math
import time
from typing import List, Optional

from otx.api.entities.train_parameters import UpdateProgressCallback
from otx.api.usecases.reporting.callback import Callback

logger = logging.getLogger(__name__)


class TimeMonitorCallback(Callback):
    """A callback to monitor the progress of training.

    Args:
        num_epoch (int): Amount of epochs
        num_train_steps (int): amount of training steps per epoch
        num_val_steps (int): amount of validation steps per epoch
        num_test_steps (int): amount of testing steps
        epoch_history (int): Amount of previous epochs to calculate average epoch time over
        step_history (int): Amount of previous steps to calculate average steps time over
        update_progress_callback (Optional[UpdateProgressCallback]): Callback to update progress
    """

    def __init__(
        self,
        num_epoch: int = 0,
        num_train_steps: int = 0,
        num_val_steps: int = 0,
        num_test_steps: int = 0,
        epoch_history: int = 5,
        step_history: int = 50,
        update_progress_callback: Optional[UpdateProgressCallback] = None,
    ):

        self.total_epochs = num_epoch
        self.train_steps = num_train_steps
        self.val_steps = num_val_steps
        self.test_steps = num_test_steps
        self.steps_per_epoch = self.train_steps + self.val_steps
        self.total_steps = math.ceil(self.steps_per_epoch * self.total_epochs + num_test_steps)
        self.current_step = 0
        self.current_epoch = 0

        # Step time calculation
        self.start_step_time = time.time()
        self.past_step_duration: List[float] = []
        self.average_step = 0
        self.step_history = step_history

        # epoch time calculation
        self.start_epoch_time = time.time()
        self.past_epoch_duration: List[float] = []
        self.average_epoch = 0
        self.epoch_history = epoch_history

        # whether model is training flag
        self.is_training = False

        self.update_progress_callback = update_progress_callback

    def on_train_batch_begin(self, batch, logs=None):
        """Set the value of current step and start the timer."""
        self.current_step += 1
        self.start_step_time = time.time()

    def on_train_batch_end(self, batch, logs=None):
        """Compute average time taken to complete a step."""
        self.__calculate_average_step()

    def is_stalling(self) -> bool:
        """Returns True if the training is stalling.

        Returns True if the current step has taken more than 30 seconds and
        at least 20x more than the average step duration
        """
        factor = 20
        min_abs_threshold = 30  # seconds
        if self.is_training and self.current_step > 2:
            step_duration = time.time() - self.start_step_time
            if step_duration > min_abs_threshold and step_duration > factor * self.average_step:
                logger.error(
                    f"Step {self.current_step} has taken {step_duration}s which is "
                    f">{min_abs_threshold}s and  {factor} times "
                    f"more than the expected {self.average_step}s"
                )
                return True
        return False

    def __calculate_average_step(self):
        """Compute average duration taken to complete a step."""
        self.past_step_duration.append(time.time() - self.start_step_time)
        if len(self.past_step_duration) > self.step_history:
            self.past_step_duration.remove(self.past_step_duration[0])
        self.average_step = sum(self.past_step_duration) / len(self.past_step_duration)

    def on_test_batch_begin(self, batch, logs):
        """Set the number of current epoch and start the timer."""
        self.current_step += 1
        self.start_step_time = time.time()

    def on_test_batch_end(self, batch, logs):
        """Compute average time taken to complete a step based on a running average of `step_history` steps."""
        self.__calculate_average_step()

    def on_train_begin(self, logs=None):
        """Sets training to true."""
        self.is_training = True

    def on_train_end(self, logs=None):
        """Handles early stopping when the total_steps is greater than the current_step."""
        # To handle cases where early stopping stops the task the progress will still be accurate
        self.current_step = self.total_steps - self.test_steps
        self.current_epoch = self.total_epochs
        self.is_training = False

    def on_epoch_begin(self, epoch, logs=None):
        """Set the number of current epoch and start the timer."""
        self.current_epoch = epoch + 1
        self.start_epoch_time = time.time()

    def on_epoch_end(self, epoch, logs=None):
        """Computes the average time taken to complete an epoch based on a running average of `epoch_history` epochs."""
        self.past_epoch_duration.append(time.time() - self.start_epoch_time)
        self._calculate_average_epoch()
        self.update_progress_callback(self.get_progress())

    def _calculate_average_epoch(self):
        if len(self.past_epoch_duration) > self.epoch_history:
            del self.past_epoch_duration[0]
        self.average_epoch = sum(self.past_epoch_duration) / len(self.past_epoch_duration)

    def get_progress(self):
        """Returns current progress as a percentage."""
        return (self.current_step / self.total_steps) * 100



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import torch
import torch.nn.functional as F
from mmcls.models.builder import HEADS
from mmcls.models.heads import LinearClsHead

from .non_linear_cls_head import NonLinearClsHead


@HEADS.register_module()
class CustomNonLinearClsHead(NonLinearClsHead):
    """Custom Nonlinear classifier head."""

    def __init__(self, *args, **kwargs):
        super(CustomNonLinearClsHead, self).__init__(*args, **kwargs)
        self.loss_type = kwargs.get("loss", dict(type="CrossEntropyLoss"))["type"]

    def loss(self, cls_score, gt_label, feature=None):
        num_samples = len(cls_score)
        losses = dict()
        # compute loss
        if self.loss_type == "IBLoss":
            loss = self.compute_loss(cls_score, gt_label, feature=feature)
        else:
            loss = self.compute_loss(cls_score, gt_label, avg_factor=num_samples)
        if self.cal_acc:
            # compute accuracy
            acc = self.compute_accuracy(cls_score, gt_label)
            assert len(acc) == len(self.topk)
            losses["accuracy"] = {f"top-{k}": a for k, a in zip(self.topk, acc)}
        losses["loss"] = loss
        return losses

    def forward_train(self, x, gt_label):
        cls_score = self.classifier(x)
        losses = self.loss(cls_score, gt_label, feature=x)
        return losses


@HEADS.register_module()
class CustomLinearClsHead(LinearClsHead):
    """Custom linear classifier head.

    Args:
        num_classes (int): Number of categories excluding the background
            category.
        in_channels (int): Number of channels in the input feature map.
        init_cfg (dict | optional): The extra init config of layers.
            Defaults to use dict(type='Normal', layer='Linear', std=0.01).
    """

    def __init__(
        self, num_classes, in_channels, init_cfg=dict(type="Normal", layer="Linear", std=0.01), *args, **kwargs
    ):
        self.loss_type = kwargs.get("loss", dict(type="CrossEntropyLoss"))["type"]
        super(CustomLinearClsHead, self).__init__(num_classes, in_channels, init_cfg=init_cfg, *args, **kwargs)

    def loss(self, cls_score, gt_label, feature=None):
        num_samples = len(cls_score)
        losses = dict()
        # compute loss
        if self.loss_type == "IBLoss":
            loss = self.compute_loss(cls_score, gt_label, feature=feature)
        else:
            loss = self.compute_loss(cls_score, gt_label, avg_factor=num_samples)
        if self.cal_acc:
            # compute accuracy
            acc = self.compute_accuracy(cls_score, gt_label)
            assert len(acc) == len(self.topk)
            losses["accuracy"] = {f"top-{k}": a for k, a in zip(self.topk, acc)}
        losses["loss"] = loss
        return losses

    def simple_test(self, img):
        """Test without augmentation."""
        cls_score = self.fc(img)
        if isinstance(cls_score, list):
            cls_score = sum(cls_score) / float(len(cls_score))
        if torch.onnx.is_in_onnx_export():
            return cls_score
        pred = F.softmax(cls_score, dim=1) if cls_score is not None else None

        return self.post_process(pred)

    def forward_train(self, x, gt_label):
        cls_score = self.fc(x)
        losses = self.loss(cls_score, gt_label, feature=x)
        return losses



"""NNCF utils."""
# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from dataclasses import dataclass, field
from typing import Any, Dict, Optional

import numpy as np
import torch

from .utils import check_nncf_is_enabled, get_nncf_version, is_nncf_enabled


@dataclass
class NNCFMetaState:
    """NNCF meta state wrapper."""

    state_to_build: Optional[Dict[str, torch.Tensor]] = field(default=None)
    data_to_build: Optional[np.ndarray] = field(default=None)
    compression_ctrl: Optional[Dict[Any, Any]] = field(default=None)

    def __repr__(self):
        """Repr."""
        out = f"{self.__class__.__name__}("
        if self.state_to_build is not None:
            out += "state_to_build='<data>', "
        if self.data_to_build is not None:
            out += "data_to_build='<data>', "
        if self.compression_ctrl is not None:
            out += "compression_ctrl='<data>', "
        if out[-2:] == ", ":
            out = out[:-2]
        out += ")"
        return out


def get_nncf_metadata():
    """Get NNCF related metadata.

    The function returns NNCF metadata that should be stored into a checkpoint.
    The metadata is used to check in wrap_nncf_model if the checkpoint should be used
    to resume NNCF training or initialize NNCF fields of NNCF-wrapped model.
    """
    check_nncf_is_enabled()
    return dict(nncf_enable_compression=True, nncf_version=get_nncf_version())


def is_state_nncf(state):
    """Check if state_dict is NNCF state_dict.

    The function uses metadata stored in a dict_state to check if the
    checkpoint was the result of trainning of NNCF-compressed model.
    See the function get_nncf_metadata above.
    """
    return bool(state.get("meta", {}).get("nncf_enable_compression", False))


def is_checkpoint_nncf(path):
    """Check if path is NNCF checkpoint.

    The function uses metadata stored in a checkpoint to check if the
    checkpoint was the result of trainning of NNCF-compressed model.
    See the function get_nncf_metadata above.
    """
    try:
        checkpoint = torch.load(path, map_location="cpu")
        return is_state_nncf(checkpoint)
    except FileNotFoundError:
        return False


def get_uncompressed_model(module):
    """Get uncompressed original model."""

    if not is_nncf_enabled():
        return module
    from nncf.torch.nncf_network import NNCFNetwork

    if isinstance(module, NNCFNetwork):
        return module.get_nncf_wrapped_model()
    return module


class AccuracyAwareLrUpdater:
    """AccuracyAwareLrUpdater."""

    def __init__(self, lr_hook):
        self._lr_hook = lr_hook
        self._lr_hook.warmup_iters = 0

    def step(self, *args, **kwargs):
        """step."""

    @property
    def base_lrs(self):
        """base_lrs."""
        return self._lr_hook.base_lr

    @base_lrs.setter
    def base_lrs(self, value):
        self._lr_hook.base_lr = value



"""NNCF utils."""
# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from collections import OrderedDict
from contextlib import contextmanager
from importlib.util import find_spec

import torch

_is_nncf_enabled = find_spec("nncf") is not None


def is_nncf_enabled():
    """is_nncf_enabled."""
    return _is_nncf_enabled


def check_nncf_is_enabled():
    """check_nncf_is_enabled."""
    if not is_nncf_enabled():
        raise RuntimeError("Tried to use NNCF, but NNCF is not installed")


def get_nncf_version():
    """get_nncf_version."""
    if not is_nncf_enabled():
        return None
    import nncf

    return nncf.__version__


def load_checkpoint(model, filename, map_location=None, strict=False):
    """Load checkpoint from a file or URI.

    Args:
        model (Module): Module to load checkpoint.
        filename (str): Either a filepath or URL or modelzoo://xxxxxxx.
        map_location (str): Same as :func:`torch.load`.
        strict (bool): Whether to allow different params for the model and
            checkpoint.

    Returns:
        dict or OrderedDict: The loaded checkpoint.
    """
    #  from nncf.torch import load_state
    from mmcv.runner import load_state_dict

    checkpoint = torch.load(filename, map_location=map_location)
    nncf_state = None
    compression_state = None
    # get state_dict from checkpoint
    if isinstance(checkpoint, OrderedDict):
        base_state = checkpoint
    elif isinstance(checkpoint, dict) and "state_dict" in checkpoint:
        if "meta" in checkpoint and "nncf_meta" in checkpoint["meta"]:
            nncf_state = checkpoint["state_dict"]
            compression_state = checkpoint["meta"]["nncf_meta"].compression_ctrl
            base_state = checkpoint["meta"]["nncf_meta"].state_to_build
        else:
            base_state = checkpoint["state_dict"]
    else:
        raise RuntimeError(f"No state_dict found in checkpoint file {filename}")
    load_state_dict(model, base_state, strict=strict)
    return compression_state, nncf_state


@contextmanager
def nullcontext():
    """Context which does nothing."""
    yield


def no_nncf_trace():
    """Wrapper for original NNCF no_nncf_trace context."""

    if is_nncf_enabled():
        from nncf.torch.dynamic_graph.context import (
            no_nncf_trace as original_no_nncf_trace,
        )

        return original_no_nncf_trace()
    return nullcontext()


def nncf_trace():
    """Trace nncf context."""
    if is_nncf_enabled():

        @contextmanager
        def _nncf_trace():
            from nncf.torch.dynamic_graph.context import get_current_context

            ctx = get_current_context()
            if ctx is not None and not ctx.is_tracing:
                ctx.enable_tracing()
                yield
                ctx.disable_tracing()
            else:
                yield

        return _nncf_trace()
    return nullcontext()


def is_in_nncf_tracing():
    """is_in_nncf_tracing."""
    if not is_nncf_enabled():
        return False

    from nncf.torch.dynamic_graph.context import get_current_context

    ctx = get_current_context()

    if ctx is None:
        return False
    return ctx.is_tracing


def is_accuracy_aware_training_set(nncf_config):
    """is_accuracy_aware_training_set."""
    if not is_nncf_enabled():
        return False
    from nncf.config.utils import is_accuracy_aware_training

    is_acc_aware_training_set = is_accuracy_aware_training(nncf_config)
    return is_acc_aware_training_set



"""This module contains the exceptions for the tasks."""


# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#


class UnrecoverableTaskException(Exception):
    """Exception for when task is in an unrecoverable state."""

    def __init__(self, message="Unrecoverable task exception"):
        # pylint: disable=W0235
        super().__init__(message)


class OOMException(UnrecoverableTaskException):
    """Exception for when task is out of memory."""

    def __init__(self, message="Out of memory exception"):
        super().__init__(message)


class TrainingStallException(UnrecoverableTaskException):
    """Exception for when training should be stalled."""

    def __init__(self, message="Training stalling exception"):
        super().__init__(message)



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from typing import Dict, List, Optional, Union

import openvino.runtime as ov
import torch
from mmdet.models.builder import NECKS
from mmdet.models.necks.yolo_neck import YOLOV3Neck

from ...mmov_model import MMOVModel
from ...parser_mixin import ParserMixin


@NECKS.register_module()
class MMOVYOLOV3Neck(YOLOV3Neck, ParserMixin):
    def __init__(
        self,
        model_path_or_model: Union[str, ov.Model],
        weight_path: Optional[str] = None,
        inputs: Optional[Union[Dict[str, Union[str, List[str]]], List[str], str]] = None,
        outputs: Optional[Union[Dict[str, Union[str, List[str]]], List[str], str]] = None,
        init_weight: bool = False,
        verify_shape: bool = True,
    ):
        super(YOLOV3Neck, self).__init__()

        self._model_path_or_model = model_path_or_model
        self._weight_path = weight_path
        self._init_weight = init_weight

        inputs, outputs = super().parse(
            model_path_or_model=model_path_or_model,
            weight_path=weight_path,
            inputs=inputs,
            outputs=outputs,
        )

        for key in inputs.keys():
            input = inputs[key]
            output = outputs[key]

            if input and output:
                self.add_module(
                    key,
                    MMOVModel(
                        model_path_or_model,
                        weight_path,
                        inputs=input,
                        outputs=output,
                        remove_normalize=False,
                        init_weight=init_weight,
                        verify_shape=verify_shape,
                    ),
                )
            else:
                self.add_module(key, torch.nn.Identity())

        self.num_scales = len([key for key in inputs.keys() if key.startswith("detect")])

    def init_weights(self, pretrained=None):
        # TODO
        pass



"""Configurable parameters for Padim anomaly task."""

# Copyright (C) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from attr import attrs

from otx.algorithms.anomaly.configs.base.configuration import BaseAnomalyConfig
from otx.algorithms.anomaly.configs.base.configuration_enums import ModelBackbone
from otx.api.configuration.elements import (
    add_parameter_group,
    selectable,
    string_attribute,
)


@attrs
class PadimAnomalyBaseConfig(BaseAnomalyConfig):
    """Configurable parameters for PADIM anomaly classification task."""

    header = string_attribute("Configuration for Padim")
    description = header

    @attrs
    class LearningParameters(BaseAnomalyConfig.LearningParameters):
        """Parameters that can be tuned using HPO."""

        header = string_attribute("Learning Parameters")
        description = header

        # Editable is set to false as WideResNet50 is very large for
        # onnx's protobuf (2gb) limit. This ends up crashing the export.
        backbone = selectable(
            default_value=ModelBackbone.RESNET18,
            header="Model Backbone",
            description="Pre-trained backbone used for feature extraction",
            editable=False,
            visible_in_ui=False,
        )

    learning_parameters = add_parameter_group(LearningParameters)



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import numpy as np
import torch
import torch.nn.functional as F
from mmcls.models.builder import LOSSES
from mmcls.models.losses import CrossEntropyLoss


@LOSSES.register_module()
class IBLoss(CrossEntropyLoss):
    def __init__(self, num_classes, start=5, alpha=1000.0, **kwargs):
        """IB Loss
        https://arxiv.org/abs/2110.02444

        Args:
            num_classes (int): Number of classes in dataset
            start (int): Epoch to start finetuning with IB loss
            alpha (float): Hyper-parameter for an adjustment for IB loss re-weighting
        """
        super(IBLoss, self).__init__(loss_weight=1.0)
        if alpha < 0:
            raise ValueError("Alpha for IB loss should be bigger than 0")
        self.alpha = alpha
        self.epsilon = 0.001
        self.num_classes = num_classes
        self.weight = None
        self._start_epoch = start
        self._cur_epoch = 0

    @property
    def cur_epoch(self):
        return self._cur_epoch

    @cur_epoch.setter
    def cur_epoch(self, epoch):
        self._cur_epoch = epoch

    def update_weight(self, cls_num_list):
        if len(cls_num_list) == 0:
            raise ValueError("Cannot compute the IB loss weight with empty cls_num_list.")
        per_cls_weights = 1.0 / np.array(cls_num_list)
        per_cls_weights = per_cls_weights / np.sum(per_cls_weights) * len(cls_num_list)
        per_cls_weights = torch.FloatTensor(per_cls_weights)
        self.weight = per_cls_weights

    def forward(self, input, target, feature):
        if self._cur_epoch < self._start_epoch:
            return super().forward(input, target)
        else:
            grads = torch.sum(torch.abs(F.softmax(input, dim=1) - F.one_hot(target, self.num_classes)), 1)
            feature = torch.sum(torch.abs(feature), 1).reshape(-1, 1)
            ib = grads * feature.reshape(-1)
            ib = self.alpha / (ib + self.epsilon)
            ce_loss = F.cross_entropy(input, target, weight=self.weight.to(input.get_device()), reduction="none")
            loss = ce_loss * ib
            return loss.mean()



"""OTX Core Data Adapter."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

# pylint: disable=too-many-return-statements
import importlib

from otx.algorithms.common.configs.training_base import TrainType
from otx.api.entities.model_template import TaskType

ADAPTERS = {
    TaskType.CLASSIFICATION: {
        "INCREMENTAL": {
            "module_name": "classification_dataset_adapter",
            "class": "ClassificationDatasetAdapter",
        }
    },
    TaskType.DETECTION: {
        "INCREMENTAL": {
            "module_name": "detection_dataset_adapter",
            "class": "DetectionDatasetAdapter",
        }
    },
    TaskType.ROTATED_DETECTION: {
        "INCREMENTAL": {
            "module_name": "detection_dataset_adapter",
            "class": "DetectionDatasetAdapter",
        }
    },
    TaskType.INSTANCE_SEGMENTATION: {
        "INCREMENTAL": {
            "module_name": "detection_dataset_adapter",
            "class": "DetectionDatasetAdapter",
        }
    },
    TaskType.SEGMENTATION: {
        "INCREMENTAL": {
            "module_name": "segmentation_dataset_adapter",
            "class": "SegmentationDatasetAdapter",
        },
        "SELFSUPERVISED": {
            "module_name": "segmentation_dataset_adapter",
            "class": "SelfSLSegmentationDatasetAdapter",
        },
    },
    TaskType.ACTION_CLASSIFICATION: {
        "INCREMENTAL": {
            "module_name": "action_dataset_adapter",
            "class": "ActionClassificationDatasetAdapter",
        }
    },
    TaskType.ACTION_DETECTION: {
        "INCREMENTAL": {
            "module_name": "action_dataset_adapter",
            "class": "ActionDetectionDatasetAdapter",
        }
    },
    TaskType.ANOMALY_CLASSIFICATION: {
        "INCREMENTAL": {
            "module_name": "anomaly_dataset_adapter",
            "class": "AnomalyClassificationDatasetAdapter",
        }
    },
    TaskType.ANOMALY_DETECTION: {
        "INCREMENTAL": {
            "module_name": "anomaly_dataset_adapter",
            "class": "AnomalyDetectionDatasetAdapter",
        }
    },
    TaskType.ANOMALY_SEGMENTATION: {
        "INCREMENTAL": {
            "module_name": "anomaly_dataset_adapter",
            "class": "AnomalySegmentationDatasetAdapter",
        }
    },
}


def get_dataset_adapter(
    task_type: TaskType,
    train_type: TrainType,
    train_data_roots: str = None,
    val_data_roots: str = None,
    test_data_roots: str = None,
    unlabeled_data_roots: str = None,
):
    """Returns a dataset class by task type.

    Args:
        task_type: A task type such as ANOMALY_CLASSIFICATION, ANOMALY_DETECTION, ANOMALY_SEGMENTATION,
            CLASSIFICATION, INSTANCE_SEGMENTATION, DETECTION, CLASSIFICATION, ROTATED_DETECTION, SEGMENTATION.
        train_type: train type such as INCREMENTAL and SELFSUPERVISED.
            SELFSUPERVISED is only supported for SEGMENTATION.
        train_data_roots: the path of data root for training data
        val_data_roots: the path of data root for validation data
        test_data_roots: the path of data root for test data
        unlabeled_data_roots: the path of data root for unlabeled data
    """

    train_type_to_be_called = TrainType.INCREMENTAL.value
    # FIXME : Hardcoded solution for self-sl for seg
    if task_type == TaskType.SEGMENTATION and train_type == TrainType.SELFSUPERVISED.value:
        train_type_to_be_called = TrainType.SELFSUPERVISED.value
    module_root = "otx.core.data.adapter."
    module = importlib.import_module(module_root + ADAPTERS[task_type][train_type_to_be_called]["module_name"])

    return getattr(module, ADAPTERS[task_type][train_type_to_be_called]["class"])(
        task_type=task_type,
        train_data_roots=train_data_roots,
        val_data_roots=val_data_roots,
        test_data_roots=test_data_roots,
        unlabeled_data_roots=unlabeled_data_roots,
    )



"""This module contains utility functions for use in defining ui rules."""
# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#


from typing import Union

from otx.api.configuration.ui_rules.types import Action, Operator


def attr_convert_operator(operator: Union[str, Operator]) -> Operator:
    """This function converts an input operator to the correct instance of the Operator Enum.

    It is used when loading an Rule element from a yaml file.
    """
    if isinstance(operator, str):
        return Operator[operator]
    return operator


def attr_convert_action(action: Union[str, Action]) -> Action:
    """This function converts an input action to the correct instance of the Action Enum.

    It is used when loading an Rule element from a yaml file.
    """
    if isinstance(action, str):
        return Action[action]
    return action



"""Unit Test for otx.algorithms.action.adapters.mmaction.heads.roi_head."""

# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import numpy as np
import pytest
import torch
from mmcv.utils import Config

from otx.algorithms.action.adapters.mmaction.models.heads.roi_head import AVARoIHead
from tests.test_suite.e2e_test_system import e2e_pytest_unit


class MockAVARoIHead(AVARoIHead):
    """Mock class for AVARoIHead."""

    def __init__(self):
        self.bbox_head = Config()
        self.test_cfg = Config()
        self.bbox_head.num_classes = 3
        self.test_cfg.action_thr = 0.5


class TestAVARoIHead:
    """Check AVARoIHead class.

    1. Check simple_test function
    <Steps>
        1. Generate sample tensor(1, 432, 32, 8, 11)
        2. Generate proposal_list: List[Tensor]
        3. Generate img_metas: List[Dict[str, Any]]
        4. Check bbox_results
            4-1. Check output lenth is fit with num_classes
            4-2. Check output has appropriate information(x1, y1, x2, y2, conf)
    """

    @pytest.fixture(autouse=True)
    def setup(self) -> None:
        self.roi_head = AVARoIHead(
            bbox_roi_extractor=dict(
                type="SingleRoIExtractor3D", roi_layer_type="RoIAlign", output_size=8, with_temporal_pool=True
            ),
            bbox_head=dict(type="BBoxHeadAVA", in_channels=432, num_classes=81, multilabel=False, dropout_ratio=0.5),
        )
        self.roi_head.test_cfg = Config()
        self.roi_head.test_cfg.action_thr = 0.5

    @e2e_pytest_unit
    def test_simple_test(self, mocker) -> None:
        """Test simple test function."""

        sample_input = torch.randn(1, 432, 32, 8, 1)
        proposal_list = [torch.Tensor([[0, 0, 10, 10]])]
        img_metas = [{"scores": np.array([1.0]), "img_shape": (256, 256)}]

        with torch.no_grad():
            out = self.roi_head.simple_test(sample_input, proposal_list, img_metas)

        assert len(out[0]) == self.roi_head.bbox_head.num_classes - 1
        assert out[0][0].shape[1] == 5

        mocker.patch(
            "otx.algorithms.action.adapters.mmaction.models.heads.roi_head.is_in_onnx_export", return_value=True
        )
        with torch.no_grad():
            bboxes, labels = self.roi_head.simple_test(sample_input, proposal_list, img_metas)
            assert isinstance(bboxes, torch.Tensor)
            assert isinstance(labels, torch.Tensor)
            assert bboxes.shape[-1] == 4
            assert labels.shape[-1] == self.roi_head.bbox_head.num_classes



"""This module contains the configuration helper functions.

These can be used to create, convert or interact with OTX configuration objects or dictionaries, yaml strings or yaml
files representing those objects.
"""
# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#


from .convert import convert
from .create import create
from .substitute import substitute_values, substitute_values_for_lifecycle
from .utils import config_to_bytes
from .validate import validate

__all__ = [
    "create",
    "config_to_bytes",
    "validate",
    "convert",
    "substitute_values",
    "substitute_values_for_lifecycle",
]



"""Convert dataset: Public dataset (Jester[RawFrames], AVA) --> Datumaro dataset (CVAT).

This script contains a lot of hardcoding to create an.xml file that Datumaro can consume.

Current Datumaro format for video (CVAT)

root
|- video_0
    |- images
        |- frames_001.png
        |- frames_002.png
    |- annotations.xml
|- video_1
    |- images
    |- annotations.xml
|- video_2

"""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

# pylint: disable=too-many-locals, c-extension-no-member, invalid-name, too-many-statements
import argparse
import csv
import os
import os.path as osp
import pathlib
import shutil
from typing import List

import cv2
from lxml import etree
from tqdm import tqdm


def generate_default_cvat_xml_fields(i, video_path, frame_list):
    """Generate default CVAT xml fields required to make multi-video CVAT format by using Jester, and AVA dataset."""
    n_frames = len(frame_list)
    annotations = etree.Element("annotations")

    version = etree.Element("version")
    version.text = "1.1"
    annotations.append(version)

    meta = etree.Element("meta")
    annotations.append(meta)

    task = etree.Element("task")
    meta.append(task)

    _id = etree.Element("id")
    _id.text = str(i)
    task.append(_id)

    name = etree.Element("name")
    name.text = f"v{i}"
    task.append(name)

    size = etree.Element("size")
    size.text = str(n_frames)
    task.append(size)

    mode = etree.Element("mode")
    mode.text = "interpolation"
    task.append(mode)

    overlap = etree.Element("overlap")
    overlap.text = "2"
    task.append(overlap)

    bugtracker = etree.Element("bugtracker")
    bugtracker.text = ""
    task.append(bugtracker)

    created = etree.Element("created")
    created.text = ""
    task.append(created)

    updated = etree.Element("updated")
    updated.text = ""
    task.append(updated)

    start_frame = etree.Element("start_frame")
    start_frame.text = "0"
    task.append(start_frame)

    stop_frame = etree.Element("stop_frame")
    stop_frame.text = str(n_frames - 1)
    task.append(stop_frame)

    frame_filter = etree.Element("frame_filter")
    frame_filter.text = "1"
    task.append(frame_filter)

    z_order = etree.Element("z_order")
    z_order.text = str(True)
    task.append(z_order)

    labels = etree.Element("labels")
    task.append(labels)

    segments = etree.Element("segments")
    segments.text = ""
    task.append(segments)

    original_size = etree.Element("original_size")
    task.append(original_size)

    sample_frame = cv2.imread(osp.join(video_path, frame_list[0]))
    original_size_width = etree.Element("width")
    original_size_width.text = str(sample_frame.shape[1])
    original_size.append(original_size_width)

    original_size_height = etree.Element("height")
    original_size_height.text = str(sample_frame.shape[0])
    original_size.append(original_size_height)

    return annotations, sample_frame.shape, labels


# classification
def convert_action_cls_dataset_to_datumaro(src_path: str, dst_path: str, ann_file: str, label_map=None):
    """Convert a public dataset to multi-video CVAT (Datumaro) format.

    Supported datasets are: Jester, HMDB51, UCF101

    Args:
        src_path (str): The path to the directory containing the video files as rawframe folder.
        dst_path (str): The path to the directory where the multi-video CVAT (Datumaro) format dataset will be saved.
        ann_file (str): The path to the file containing the annotations for the videos.
        label_map (optional): The path to the file containing the mapping between class IDs and class names.

    Returns:
        None

    Examples:
        src_path = "./data/hmdb51/rawframes"
        dst_path = "./data/hmdb51/CVAT/train"
        ann_file = "./data/hmdb51/hmdb51_train_split_1_rawframes.txt"
        label_map = "./data/hmdb51/label_map.txt"
        convert_public_dataset_to_datumaro(src_path, dst_path, ann_file, label_map=label_map)
    """

    # Load the annotations file. Annotation is supposed as whitespace separated format: Video_name num_frames class_idx
    # Ex) kiss/American_History_X_kiss_h_cm_np2_le_goo_40 69 22
    with open(ann_file, "r", encoding="utf-8") as anns:
        pathlib.Path(osp.join(dst_path)).mkdir(parents=True, exist_ok=True)

        lines = anns.readlines()
        for i, line in tqdm(enumerate(lines), total=len(lines)):
            if line[0] == "#":
                continue

            # Parse the video directory and class ID from the annotations file
            video_dir, _, class_idx = line[:-1].split(" ")

            if label_map is not None:
                # Load label mapping file : brush_hair\ncarwheel\ncatch, ...
                with open(label_map, "r", encoding="utf-8") as f:
                    label_names = f.read().splitlines()
                label_mapping_dict = {str(idx): label_name for idx, label_name in enumerate(label_names)}
                class_name = label_mapping_dict[class_idx]
            else:
                class_name = class_idx

            # Prepare the output directories and file names
            video_path = osp.join(src_path, video_dir)
            video_name = f"Video_{i}"
            images_dir = osp.join(dst_path, f"{video_name}/images")

            # List the frames in the video and sort them
            frame_list = os.listdir(video_path)
            frame_list.sort()

            # Generate default CVAT XML fields for the video annotation
            annotations, img_shape, labels = generate_default_cvat_xml_fields(i, video_path, frame_list)

            # Add the video label to the annotations
            label = etree.Element("label")
            labels.append(label)

            name = etree.Element("name")
            name.text = class_name
            label.append(name)

            attributes = etree.Element("attributes")
            attributes.text = ""
            label.append(attributes)

            # Copy the video frames to the output directory and create the image tags in the annotations
            for j, frame in enumerate(frame_list):
                if not osp.exists(images_dir):
                    os.makedirs(images_dir, exist_ok=True)

                image_name = f"{j+1:05d}.jpg"
                shutil.copy(osp.join(video_path, frame), osp.join(images_dir, image_name))
                image = etree.Element(
                    "image", id=str(j), name=image_name, width=str(img_shape[1]), height=str(img_shape[0])
                )
                tag = etree.Element("tag", label=class_name, source="manual")
                tag.text = ""
                image.append(tag)
                annotations.append(image)

            et = etree.ElementTree(annotations)
            et.write(
                osp.join(dst_path, f"{video_name}/annotations.xml"),
                pretty_print=True,
                xml_declaration=True,
                encoding="utf-8",
            )


def convert_ava_dataset_to_datumaro(src_path: str, dst_path: str, ann_file: str):
    """Convert AVA dataset to multi-video CVAT (Datumaro) format.

    Args:
        src_path (str): The path to the directory containing the video files as rawframe folder.
        dst_path (str): The path to the directory where the multi-video CVAT (Datumaro) format dataset will be saved.
        ann_file (str): The path to the file containing the annotations for the videos.

    Returns:
        None

    Examples:
        src_path = "./data/ava_dataset/frames"
        dst_path = "./data/ava_dataset/CVAT/train"
        ann_file = "./data/ava_dataset/annotations/train.csv"
        convert_ava_dataset_to_datumaro(src_path, dst_path, ann_file)
    """

    video_dict = {}
    video_idx = 0
    annot_info = read_ava_csv(ann_file)
    video_dirs = os.listdir(src_path)
    video_dirs.sort()
    for video_dir in video_dirs:
        if video_dir not in annot_info:
            continue
        video_path = osp.join(src_path, video_dir)
        frame_list = os.listdir(video_path)
        frame_list.sort()
        if video_dir not in video_dict:
            annotations, _, labels = generate_default_cvat_xml_fields(video_idx, video_path, frame_list)
            label_list: List[str] = []
            video_dict[video_dir] = {"annotations": annotations, "labels": labels, "label_list": label_list}
        for frame in frame_list:
            frame_idx = int(frame.split(".")[0].split("_")[-1])
            if frame_idx in annot_info[video_dir]:
                annots = annot_info[video_dir][frame_idx]
            else:
                annots = [[0, 0, 0, 0, "EmptyFrame"]]

            for annot in annots:
                bboxes = annot[:4]
                class_idx = annot[4]
                track = etree.Element("track", id=str(video_idx), label=str(class_idx))
                if class_idx not in video_dict[video_dir]["label_list"]:
                    video_dict[video_dir]["label_list"].append(class_idx)

                    label = etree.Element("label")
                    video_dict[video_dir]["labels"].append(label)

                    name = etree.Element("name")
                    name.text = str(class_idx)
                    label.append(name)

                    attributes = etree.Element("attributes")
                    attributes.text = ""
                    label.append(attributes)
                box = etree.Element(
                    "box",
                    frame=str(frame_idx),
                    outside="0",  # not used
                    occluded="0",  # not used
                    xtl=str(bboxes[0]),
                    ytl=str(bboxes[1]),
                    xbr=str(bboxes[2]),
                    ybr=str(bboxes[3]),
                    z_order="0",
                    label=str(class_idx),
                )
                box.text = ""
                track.append(box)
                video_dict[video_dir]["annotations"].append(track)

    for (video_dir, info) in video_dict.items():
        video_path = osp.join(src_path, video_dir)
        shutil.copytree(video_path, osp.join(dst_path, f"{video_dir}/images"), copy_function=rename_and_copy)
        et = etree.ElementTree(info["annotations"])
        et.write(
            osp.join(dst_path, f"{video_dir}/annotations.xml"),
            pretty_print=True,
            xml_declaration=True,
            encoding="utf-8",
        )


def rename_and_copy(_src, _dst):
    """Change frame name to cvat format."""
    img_name = _dst.rsplit("/", maxsplit=1)[-1]
    # FIXME This only support AVA dataset name
    if "_" in img_name:
        frame_index = int(img_name.split(".")[0].split("_")[-1])
    else:
        frame_index = int(img_name.split(".")[0])
    new_img_name = f"frame_{frame_index:06d}.png"
    _dst = _dst.replace(img_name, new_img_name)
    shutil.copy2(_src, _dst)


def read_ava_csv(csv_path):
    """Read ava format annotation csv file."""
    annot_info = {}  # {video_id: {frame_idx: [annot0, annot1, ..., annotN]}}
    with open(csv_path, "r", encoding="utf-8") as csv_file:
        csv_reader = csv.reader(csv_file, delimiter=",")
        for line in csv_reader:
            video_id, frame_idx, bboxes, class_idx = line[0], line[1], line[2:6], line[6]
            frame_idx = int(frame_idx)
            bboxes.append(class_idx)
            if video_id in annot_info:
                if frame_idx in annot_info[video_id]:
                    annot_info[video_id][frame_idx].append(bboxes)
                else:
                    annot_info[video_id][frame_idx] = [bboxes]
            else:
                annot_info[video_id] = {frame_idx: [bboxes]}
    return annot_info


def parse_args():
    """Parses command line arguments."""
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument("--task", required=True, type=str)
    parser.add_argument("--src_path", required=True, type=str)
    parser.add_argument("--dst_path", required=True, type=str)
    parser.add_argument("--ann_file", required=True, type=str)
    parser.add_argument("--label_map", type=str, default=None)
    return parser.parse_args()


def main():
    """Main function."""
    args = parse_args()
    if args.task == "action_classification":
        convert_action_cls_dataset_to_datumaro(args.src_path, args.dst_path, args.ann_file, args.label_map)
    elif args.task == "action_detection":
        convert_ava_dataset_to_datumaro(args.src_path, args.dst_path, args.ann_file)


if __name__ == "__main__":
    main()



"""Base MMDataset for Detection Task."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import tempfile
from collections import OrderedDict
from copy import copy
from typing import Any, Dict, List, Sequence, Tuple, Union

import numpy as np
from mmcv import Config
from mmcv.utils import print_log
from mmdet.core import PolygonMasks, eval_map
from mmdet.datasets.builder import DATASETS, build_dataset
from mmdet.datasets.custom import CustomDataset
from mmdet.datasets.pipelines import Compose

from otx.algorithms.common.utils.data import get_old_new_img_indices
from otx.algorithms.detection.adapters.mmdet.evaluation import eval_segm
from otx.api.entities.dataset_item import DatasetItemEntity
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.label import Domain, LabelEntity
from otx.api.utils.argument_checks import (
    DatasetParamTypeCheck,
    check_input_parameters_type,
)
from otx.api.utils.shape_factory import ShapeFactory

from .tiling import Tile


# pylint: disable=invalid-name, too-many-locals, too-many-instance-attributes, super-init-not-called
@check_input_parameters_type()
def get_annotation_mmdet_format(
    dataset_item: DatasetItemEntity,
    labels: List[LabelEntity],
    domain: Domain,
    min_size: int = -1,
) -> dict:
    """Function to convert a OTX annotation to mmdetection format.

    This is used both in the OTXDataset class defined in
    this file as in the custom pipeline element 'LoadAnnotationFromOTXDataset'

    :param dataset_item: DatasetItem for which to get annotations
    :param labels: List of labels that are used in the task
    :return dict: annotation information dict in mmdet format
    """
    width, height = dataset_item.width, dataset_item.height

    # load annotations for item
    gt_bboxes = []
    gt_labels = []
    gt_polygons = []

    label_idx = {label.id: i for i, label in enumerate(labels)}

    for annotation in dataset_item.get_annotations(labels=labels, include_empty=False):

        box = ShapeFactory.shape_as_rectangle(annotation.shape)

        if min(box.width * width, box.height * height) < min_size:
            continue

        class_indices = [
            label_idx[label.id] for label in annotation.get_labels(include_empty=False) if label.domain == domain
        ]

        n = len(class_indices)
        gt_bboxes.extend([[box.x1 * width, box.y1 * height, box.x2 * width, box.y2 * height] for _ in range(n)])
        if domain != Domain.DETECTION:
            polygon = ShapeFactory.shape_as_polygon(annotation.shape)
            polygon = np.array([p for point in polygon.points for p in [point.x * width, point.y * height]])
            gt_polygons.extend([[polygon] for _ in range(n)])
        gt_labels.extend(class_indices)

    if len(gt_bboxes) > 0:
        ann_info = dict(
            bboxes=np.array(gt_bboxes, dtype=np.float32).reshape(-1, 4),
            labels=np.array(gt_labels, dtype=int),
            masks=PolygonMasks(gt_polygons, height=height, width=width) if gt_polygons else [],
        )
    else:
        ann_info = dict(
            bboxes=np.zeros((0, 4), dtype=np.float32),
            labels=np.array([], dtype=int),
            masks=[],
        )
    return ann_info


@DATASETS.register_module()
class OTXDetDataset(CustomDataset):
    """Wrapper that allows using a OTX dataset to train mmdetection models.

    This wrapper is not based on the filesystem,
    but instead loads the items here directly from the OTX DatasetEntity object.

    The wrapper overwrites some methods of the CustomDataset class: prepare_train_img, prepare_test_img and prepipeline
    Naming of certain attributes might seem a bit peculiar but this is due to the conventions set in CustomDataset. For
    instance, CustomDatasets expects the dataset items to be stored in the attribute data_infos, which is why it is
    named like that and not dataset_items.

    """

    class _DataInfoProxy:
        """This class is intended to be a wrapper to use it in CustomDataset-derived class as `self.data_infos`.

        Instead of using list `data_infos` as in CustomDataset, our implementation of dataset OTXDataset
        uses this proxy class with overriden __len__ and __getitem__; this proxy class
        forwards data access operations to otx_dataset and converts the dataset items to the view
        convenient for mmdetection.
        """

        def __init__(self, otx_dataset, labels):
            self.otx_dataset = otx_dataset
            self.labels = labels
            self.label_idx = {label.id: i for i, label in enumerate(labels)}

        def __len__(self):
            return len(self.otx_dataset)

        def __getitem__(self, index):
            """Prepare a dict 'data_info' that is expected by the mmdet pipeline to handle images and annotations.

            :return data_info: dictionary that contains the image and image metadata, as well as the labels of
            the objects in the image
            """

            dataset = self.otx_dataset
            item = dataset[index]
            ignored_labels = np.array([self.label_idx[lbs.id] for lbs in item.ignored_labels])

            height, width = item.height, item.width

            data_info = dict(
                dataset_item=item,
                width=width,
                height=height,
                index=index,
                ann_info=dict(label_list=self.labels),
                ignored_labels=ignored_labels,
            )

            return data_info

    @check_input_parameters_type({"otx_dataset": DatasetParamTypeCheck})
    def __init__(
        self,
        otx_dataset: DatasetEntity,
        labels: List[LabelEntity],
        pipeline: Sequence[dict],
        domain: Domain,
        test_mode: bool = False,
        **kwargs,
    ):
        dataset_cfg = kwargs.copy()
        _ = dataset_cfg.pop("org_type", None)
        new_classes = dataset_cfg.pop("new_classes", [])
        self.otx_dataset = otx_dataset
        self.labels = labels
        self.CLASSES = list(label.name for label in labels)
        self.domain = domain
        self.test_mode = test_mode

        # Instead of using list data_infos as in CustomDataset, this implementation of dataset
        # uses a proxy class with overriden __len__ and __getitem__; this proxy class
        # forwards data access operations to otx_dataset.
        # Note that list `data_infos` cannot be used here, since OTX dataset class does not have interface to
        # get only annotation of a data item, so we would load the whole data item (including image)
        # even if we need only checking aspect ratio of the image; due to it
        # this implementation of dataset does not uses such tricks as skipping images with wrong aspect ratios or
        # small image size, since otherwise reading the whole dataset during initialization will be required.
        self.data_infos = OTXDetDataset._DataInfoProxy(otx_dataset, labels)

        self.proposals = None  # Attribute expected by mmdet but not used for OTX datasets

        if not test_mode:
            self._set_group_flag()
            self.img_indices = get_old_new_img_indices(self.labels, new_classes, self.otx_dataset)

        self.pipeline = Compose(pipeline)

    def _set_group_flag(self):
        """Set flag for grouping images.

        Originally, in Custom dataset, images with aspect ratio greater than 1 will be set as group 1,
        otherwise group 0.
        This implementation will set group 0 for every image.
        """
        self.flag = np.zeros(len(self), dtype=np.uint8)

    def _rand_another(self, idx):
        _ = idx
        return np.random.choice(len(self))

    @check_input_parameters_type()
    def prepare_train_img(self, idx: int) -> dict:
        """Get training data and annotations after pipeline.

        :param idx: int, Index of data.
        :return dict: Training data and annotation after pipeline with new keys introduced by pipeline.
        """
        item = copy(self.data_infos[idx])  # Copying dict(), not contents
        self.pre_pipeline(item)
        return self.pipeline(item)

    @check_input_parameters_type()
    def prepare_test_img(self, idx: int) -> dict:
        """Get testing data after pipeline.

        :param idx: int, Index of data.
        :return dict: Testing data after pipeline with new keys introduced by pipeline.
        """
        item = copy(self.data_infos[idx])  # Copying dict(), not contents
        self.pre_pipeline(item)
        return self.pipeline(item)

    @staticmethod
    @check_input_parameters_type()
    def pre_pipeline(results: Dict[str, Any]):
        """Prepare results dict for pipeline. Add expected keys to the dict."""
        results["bbox_fields"] = []
        results["mask_fields"] = []
        results["seg_fields"] = []

    @check_input_parameters_type()
    def get_ann_info(self, idx: int):
        """This method is used for evaluation of predictions.

        The CustomDataset class implements a method
        CustomDataset.evaluate, which uses the class method get_ann_info to retrieve annotations.

        :param idx: index of the dataset item for which to get the annotations
        :return ann_info: dict that contains the coordinates of the bboxes and their corresponding labels
        """
        dataset_item = self.otx_dataset[idx]
        labels = self.labels
        return get_annotation_mmdet_format(dataset_item, labels, self.domain)

    def evaluate(  # pylint: disable=too-many-branches
        self,
        results,
        metric="mAP",
        logger=None,
        proposal_nums=(100, 300, 1000),
        iou_thr=0.5,
        scale_ranges=None,
    ):
        """Evaluate the dataset.

        Args:
            results (list): Testing results of the dataset.
            metric (str | list[str]): Metrics to be evaluated.
            logger (logging.Logger | None | str): Logger used for printing
                related information during evaluation. Default: None.
            proposal_nums (Sequence[int]): Proposal number used for evaluating
                recalls, such as recall@100, recall@1000.
                Default: (100, 300, 1000).
            iou_thr (float | list[float]): IoU threshold. Default: 0.5.
            scale_ranges (list[tuple] | None): Scale ranges for evaluating mAP.
                Default: None.
        """
        metrics = metric if isinstance(metric, list) else [metric]
        allowed_metrics = ["mAP"]
        eval_results = OrderedDict()
        for metric in metrics:  # pylint: disable=redefined-argument-from-local
            if metric not in allowed_metrics:
                raise KeyError(f"metric {metric} is not supported")
            annotations = [self.get_ann_info(i) for i in range(len(self))]
            iou_thrs = [iou_thr] if isinstance(iou_thr, float) else iou_thr
            if metric == "mAP":
                assert isinstance(iou_thrs, list)
                mean_aps = []
                for iou_thr in iou_thrs:  # pylint: disable=redefined-argument-from-local
                    print_log(f'\n{"-" * 15}iou_thr: {iou_thr}{"-" * 15}')
                    if isinstance(results[0], tuple):
                        mean_ap, _ = eval_segm(
                            results,
                            annotations,
                            iou_thr=iou_thr,
                            dataset=self.CLASSES,
                            logger=logger,
                            metric=metric,
                        )
                    else:
                        mean_ap, _ = eval_map(
                            results,
                            annotations,
                            scale_ranges=scale_ranges,
                            iou_thr=iou_thr,
                            dataset=self.CLASSES,
                            logger=logger,
                        )
                    mean_aps.append(mean_ap)
                    eval_results[f"AP{int(iou_thr * 100):02d}"] = round(mean_ap, 3)
                eval_results["mAP"] = sum(mean_aps) / len(mean_aps)
        return eval_results


# pylint: disable=too-many-arguments
@DATASETS.register_module()
class ImageTilingDataset:
    """A wrapper of tiling dataset.

    Suitable for training small object dataset. This wrapper composed of `Tile`
    that crops an image into tiles and merges tile-level predictions to
    image-level prediction for evaluation.

    Args:
        dataset (Config): The dataset to be tiled.
        pipeline (List): Sequence of transform object or
            config dict to be composed.
        tile_size (int): the length of side of each tile
        min_area_ratio (float, optional): The minimum overlap area ratio
            between a tiled image and its annotations. Ground-truth box is
            discarded if the overlap area is less than this value.
            Defaults to 0.8.
        overlap_ratio (float, optional): ratio of each tile to overlap with
            each of the tiles in its 4-neighborhood. Defaults to 0.2.
        iou_threshold (float, optional): IoU threshold to be used to suppress
            boxes in tiles' overlap areas. Defaults to 0.45.
        max_per_img (int, optional): if there are more than max_per_img bboxes
            after NMS, only top max_per_img will be kept. Defaults to 200.
    """

    def __init__(
        self,
        dataset: Config,
        pipeline: List[dict],
        tile_size: int,
        min_area_ratio=0.8,
        overlap_ratio=0.2,
        iou_threshold=0.45,
        max_per_img=200,
        filter_empty_gt=True,
        test_mode=False,
    ):
        self.dataset = build_dataset(dataset)
        self.CLASSES = self.dataset.CLASSES
        self.tmp_dir = tempfile.TemporaryDirectory()  # pylint: disable=consider-using-with

        self.tile_dataset = Tile(
            self.dataset,
            pipeline,
            tmp_dir=self.tmp_dir,
            tile_size=tile_size,
            overlap=overlap_ratio,
            min_area_ratio=min_area_ratio,
            iou_threshold=iou_threshold,
            max_per_img=max_per_img,
            filter_empty_gt=False if test_mode else filter_empty_gt,
        )
        self.flag = np.zeros(len(self), dtype=np.uint8)
        self.pipeline = Compose(pipeline)
        self.test_mode = test_mode
        self.num_samples = len(self.dataset)  # number of original samples
        self.merged_results: Union[List[Tuple[np.ndarray, list]], List[np.ndarray]] = []

    def __len__(self) -> int:
        """Get the length of the dataset."""
        return len(self.tile_dataset)

    def __getitem__(self, idx: int) -> Dict:
        """Get training/test tile.

        Args:
            idx (int): Index of data.

        Returns:
            dict: Training/test data (with annotation if `test_mode` is set
            True).
        """
        return self.pipeline(self.tile_dataset[idx])

    def evaluate(self, results, **kwargs) -> Dict[str, float]:
        """Evaluation on Tile dataset.

        Args:
            results (list[list | tuple]): Testing results of the dataset.

        Returns:
            dict[str, float]: evaluation metric.
        """
        self.merged_results = self.tile_dataset.merge(results)
        return self.dataset.evaluate(self.merged_results, **kwargs)

    def merge(self, results) -> Union[List[Tuple[np.ndarray, list]], List[np.ndarray]]:
        """Merge tile-level results to image-level results.

        Args:
            results: tile-level results.

        Returns:
            merged_results (list[list | tuple]): Merged results of the dataset.
        """
        self.merged_results = self.tile_dataset.merge(results)
        return self.merged_results

    def __del__(self):
        """Delete the temporary directory when the object is deleted."""
        if getattr(self, "tmp_dir", False):
            self.tmp_dir.cleanup()



"""This module contains interface for performance providers."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#


import abc

from otx.api.entities.metrics import Performance


class IPerformanceProvider(metaclass=abc.ABCMeta):
    """Interface for performance provider.

    TODO: subject for refactoring.
    """

    @abc.abstractmethod
    def get_performance(self) -> Performance:
        """Returns the computed performance."""
        raise NotImplementedError



"""Base Class for Dataset Adapter."""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

# pylint: disable=invalid-name, too-many-locals, no-member, too-many-instance-attributes, unused-argument

import abc
from abc import abstractmethod
from typing import Any, Dict, List, Optional, Union

import datumaro
from datumaro.components.annotation import Annotation as DatumaroAnnotation
from datumaro.components.annotation import AnnotationType as DatumaroAnnotationType
from datumaro.components.annotation import Categories as DatumaroCategories
from datumaro.components.dataset import Dataset as DatumaroDataset
from datumaro.components.dataset import DatasetSubset as DatumaroDatasetSubset

from otx.api.entities.annotation import (
    Annotation,
    AnnotationSceneEntity,
    AnnotationSceneKind,
    NullAnnotationSceneEntity,
)
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.id import ID
from otx.api.entities.label import LabelEntity
from otx.api.entities.label_schema import LabelGroup, LabelGroupType, LabelSchemaEntity
from otx.api.entities.model_template import TaskType
from otx.api.entities.scored_label import ScoredLabel
from otx.api.entities.shapes.polygon import Point, Polygon
from otx.api.entities.shapes.rectangle import Rectangle
from otx.api.entities.subset import Subset


class BaseDatasetAdapter(metaclass=abc.ABCMeta):
    """Base dataset adapter for all of downstream tasks to use Datumaro.

    Mainly, BaseDatasetAdapter detect and import the dataset by using the function implemented in Datumaro.
    And it could prepare common variable, function (EmptyLabelSchema, LabelSchema, ..) commonly consumed under all tasks

    Args:
        task_type [TaskType]: type of the task
        train_data_roots (Optional[str]): Path for training data
        val_data_roots (Optional[str]): Path for validation data
        test_data_roots (Optional[str]): Path for test data
        unlabeled_data_roots (Optional[str]): Path for unlabeled data

    Since all adapters can be used for training and validation,
    the default value of train/val/test_data_roots was set to None.

    i.e)
    For the training/validation phase, test_data_roots is not used.
    For the test phase, train_data_roots and val_data_root are not used.
    """

    def __init__(
        self,
        task_type: TaskType,
        train_data_roots: Optional[str] = None,
        val_data_roots: Optional[str] = None,
        test_data_roots: Optional[str] = None,
        unlabeled_data_roots: Optional[str] = None,
    ):
        self.task_type = task_type
        self.domain = task_type.domain
        self.data_type: str
        self.is_train_phase: bool

        self.dataset = self._import_dataset(
            train_data_roots=train_data_roots,
            val_data_roots=val_data_roots,
            test_data_roots=test_data_roots,
            unlabeled_data_roots=unlabeled_data_roots,
        )

        self.category_items: Dict[DatumaroAnnotationType, DatumaroCategories]
        self.label_groups: List[str]
        self.label_entities: List[LabelEntity]
        self.label_schema: LabelSchemaEntity

    def _import_dataset(
        self,
        train_data_roots: Optional[str] = None,
        val_data_roots: Optional[str] = None,
        test_data_roots: Optional[str] = None,
        unlabeled_data_roots: Optional[str] = None,
    ) -> Dict[Subset, DatumaroDataset]:
        """Import dataset by using Datumaro.import_from() method.

        Args:
            train_data_roots (Optional[str]): Path for training data
            val_data_roots (Optional[str]): Path for validation data
            test_data_roots (Optional[str]): Path for test data
            unlabeled_data_roots (Optional[str]): Path for unlabeled data

        Returns:
            DatumaroDataset: Datumaro Dataset
        """
        dataset = {}
        if train_data_roots is None and test_data_roots is None:
            raise ValueError("At least 1 data_root is needed to train/test.")

        # Construct dataset for training, validation, testing, unlabeled
        if train_data_roots is not None:
            # Find self.data_type and task_type
            self.data_type_candidates = self._detect_dataset_format(path=train_data_roots)
            self.data_type = self._select_data_type(self.data_type_candidates)

            train_dataset = DatumaroDataset.import_from(train_data_roots, format=self.data_type)

            # Prepare subsets by using Datumaro dataset
            dataset[Subset.TRAINING] = self._get_subset_data("train", train_dataset)
            self.is_train_phase = True

            # If validation is manually defined --> set the validation data according to user's input
            if val_data_roots:
                val_data_candidates = self._detect_dataset_format(path=val_data_roots)
                val_data_type = self._select_data_type(val_data_candidates)
                val_dataset = DatumaroDataset.import_from(val_data_roots, format=val_data_type)
                dataset[Subset.VALIDATION] = self._get_subset_data("val", val_dataset)
            else:
                if "val" in train_dataset.subsets():
                    dataset[Subset.VALIDATION] = self._get_subset_data("val", train_dataset)

        if test_data_roots is not None and train_data_roots is None:
            self.data_type_candidates = self._detect_dataset_format(path=test_data_roots)
            self.data_type = self._select_data_type(self.data_type_candidates)
            test_dataset = DatumaroDataset.import_from(test_data_roots, format=self.data_type)
            dataset[Subset.TESTING] = self._get_subset_data("test", test_dataset)
            self.is_train_phase = False

        if unlabeled_data_roots is not None:
            dataset[Subset.UNLABELED] = DatumaroDataset.import_from(unlabeled_data_roots, format="image_dir")

        return dataset

    @abstractmethod
    def get_otx_dataset(self) -> DatasetEntity:
        """Get DatasetEntity."""
        raise NotImplementedError

    def get_label_schema(self) -> LabelSchemaEntity:
        """Get Label Schema."""
        return self._generate_default_label_schema(self.label_entities)

    def _get_subset_data(self, subset: str, dataset: DatumaroDataset) -> DatumaroDatasetSubset:
        """Get subset dataset according to subset."""
        for k, v in dataset.subsets().items():
            if subset in k or "default" in k:
                return v
            if subset == "test" and "val" in k:
                return v

        raise ValueError("Can't find proper dataset.")

    def _detect_dataset_format(self, path: str) -> str:
        """Detect dataset format (ImageNet, COCO, ...)."""
        return datumaro.Environment().detect_dataset(path=path)

    def _generate_empty_label_entity(self) -> LabelGroup:
        """Generate Empty Label Group for H-label, Multi-label Classification."""
        empty_label = LabelEntity(name="Empty label", is_empty=True, domain=self.domain)
        empty_group = LabelGroup(name="empty", labels=[empty_label], group_type=LabelGroupType.EMPTY_LABEL)
        return empty_group

    def _generate_default_label_schema(self, label_entities: List[LabelEntity]) -> LabelSchemaEntity:
        """Generate Default Label Schema for Multi-class Classification, Detecion, Etc."""
        label_schema = LabelSchemaEntity()
        main_group = LabelGroup(
            name="labels",
            labels=label_entities,
            group_type=LabelGroupType.EXCLUSIVE,
        )
        label_schema.add_group(main_group)
        return label_schema

    def _prepare_label_information(
        self,
        datumaro_dataset: Dict[Subset, DatumaroDataset],
    ) -> Dict[str, Any]:
        # Get datumaro category information
        if self.is_train_phase:
            label_categories_list = (
                datumaro_dataset[Subset.TRAINING].categories().get(DatumaroAnnotationType.label, None)
            )
        else:
            label_categories_list = (
                datumaro_dataset[Subset.TESTING].categories().get(DatumaroAnnotationType.label, None)
            )
        category_items = label_categories_list.items
        label_groups = label_categories_list.label_groups

        # LabelEntities
        label_entities = [
            LabelEntity(name=class_name.name, domain=self.domain, is_empty=False, id=ID(i))
            for i, class_name in enumerate(category_items)
        ]

        return {"category_items": category_items, "label_groups": label_groups, "label_entities": label_entities}

    def _is_normal_polygon(self, annotation: DatumaroAnnotationType.polygon) -> bool:
        """To filter out the abnormal polygon."""
        x_points = [annotation.points[i] for i in range(0, len(annotation.points), 2)]
        y_points = [annotation.points[i + 1] for i in range(0, len(annotation.points), 2)]
        return min(x_points) < max(x_points) and min(y_points) < max(y_points)

    def _is_normal_bbox(self, x1: float, y1: float, x2: float, y2: float) -> bool:
        """To filter out the abrnormal bbox."""
        return x1 < x2 and y1 < y2

    def _select_data_type(self, data_candidates: Union[List[str], str]) -> str:
        """Select specific type among candidates.

        Args:
            data_candidates (Union[List[str], str]): Type candidates made by Datumaro.Environment().detect_dataset()

        Returns:
            str: Selected data type
        """
        return data_candidates[0]

    def _get_ann_scene_entity(self, shapes: List[Annotation]) -> AnnotationSceneEntity:
        annotation_scene: Optional[AnnotationSceneEntity] = None
        if len(shapes) == 0:
            annotation_scene = NullAnnotationSceneEntity()
        else:
            annotation_scene = AnnotationSceneEntity(kind=AnnotationSceneKind.ANNOTATION, annotations=shapes)
        return annotation_scene

    def _get_label_entity(self, annotation: DatumaroAnnotation) -> Annotation:
        """Get label entity."""
        return Annotation(
            Rectangle.generate_full_box(), labels=[ScoredLabel(label=self.label_entities[annotation.label])]
        )

    def _get_normalized_bbox_entity(self, annotation: DatumaroAnnotation, width: int, height: int) -> Annotation:
        """Get bbox entity w/ normalization."""
        x1, y1, x2, y2 = annotation.points
        return Annotation(
            Rectangle(
                x1=x1 / width,
                y1=y1 / height,
                x2=x2 / width,
                y2=y2 / height,
            ),
            labels=[ScoredLabel(label=self.label_entities[annotation.label])],
        )

    def _get_original_bbox_entity(self, annotation: DatumaroAnnotation) -> Annotation:
        """Get bbox entity w/o normalization."""
        return Annotation(
            Rectangle(
                x1=annotation.points[0],
                y1=annotation.points[1],
                x2=annotation.points[2],
                y2=annotation.points[3],
            ),
            labels=[ScoredLabel(label=self.label_entities[annotation.label])],
        )

    def _get_polygon_entity(self, annotation: DatumaroAnnotation, width: int, height: int) -> Annotation:
        """Get polygon entity."""
        return Annotation(
            Polygon(
                points=[
                    Point(x=annotation.points[i] / width, y=annotation.points[i + 1] / height)
                    for i in range(0, len(annotation.points), 2)
                ]
            ),
            labels=[ScoredLabel(label=self.label_entities[annotation.label])],
        )

    def remove_unused_label_entities(self, used_labels: List):
        """Remove unused label from label entities.

        Because label entities will be used to make Label Schema,
        If there is unused label in Label Schema, it will hurts the model performance.
        So, remove the unused label from label entities.

        Args:
            used_labels (List): list for index of used label
        """
        clean_label_entities = []
        for used_label in used_labels:
            clean_label_entities.append(self.label_entities[used_label])
        self.label_entities = clean_label_entities



"""Test helpers for MPA."""
import torch
import torch.nn as nn


def generate_random_torch_image(batch=1, width=3, height=3, channels=3, channel_last=False):
    """Generate random torch tensor image.

    Args:
        batch (int, optional): A size of batch. Defaults to 1.
        width (int, optional): the image width. Defaults to 224.
        height (int, optional): the image height. Defaults to 224.
        channels (int, optional): the image channel. Defaults to 3.
        channel_last (bool, optional): if this is True, image shape will follow BHWC. Defaults to True.

    Returns:
        _type_: _description_
    """
    if channel_last is False:
        img = torch.rand(batch, channels, height, width)
    else:
        img = torch.rand(batch, height, width, channels)
    return img


def generate_toy_cnn_model(in_channels=3, mid_channels=3, out_channels=3):
    return nn.Sequential(
        nn.Conv2d(in_channels, mid_channels, (1, 1)), nn.BatchNorm2d(mid_channels), nn.AdaptiveAvgPool2d((1, 1))
    )


def generate_toy_head(in_features, out_features):
    return nn.Linear(in_features, out_features)



# Copyright (C) 2021-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import numpy as np
import pytest

from otx.algorithms.classification.configs.base import ClassificationConfig
from otx.algorithms.classification.tasks.openvino import (
    ClassificationOpenVINOInferencer,
    ClassificationOpenVINOTask,
    OTXOpenVinoDataLoader,
)
from otx.api.configuration.configurable_parameters import ConfigurableParameters
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.label_schema import LabelSchemaEntity
from otx.api.entities.model import ModelConfiguration, ModelEntity
from otx.api.entities.resultset import ResultSetEntity
from otx.api.usecases.tasks.interfaces.optimization_interface import OptimizationType
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.api.parameters_validation.validation_helper import (
    check_value_error_exception_raised,
)


def model():
    model_configuration = ModelConfiguration(
        configurable_parameters=ConfigurableParameters(header="header", description="description"),
        label_schema=LabelSchemaEntity(),
    )
    return ModelEntity(train_dataset=DatasetEntity(), configuration=model_configuration)


class MockOpenVinoTask(ClassificationOpenVINOTask):
    def __init__(self):
        pass


class MockOpenVinoInferencer(ClassificationOpenVINOInferencer):
    def __init__(self):
        pass


class TestClassificationOpenVINOTaskInputParamsValidation:
    @e2e_pytest_unit
    def test_openvino_task_init_params_validation(self):
        """
        <b>Description:</b>
        Check ClassificationOpenVINOTask object initialization parameters validation

        <b>Input data:</b>
        "task_environment" non-TaskEnvironment object

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        ClassificationOpenVINOTask object initialization parameter
        """
        with pytest.raises(ValueError):
            ClassificationOpenVINOTask(task_environment="unexpected string")  # type: ignore

    @e2e_pytest_unit
    def test_openvino_task_infer_params_validation(self):
        """
        <b>Description:</b>
        Check ClassificationOpenVINOTask object "infer" method input parameters validation

        <b>Input data:</b>
        ClassificationOpenVINOTask object. "infer" method unexpected-type input parameters

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "infer" method
        """
        task = MockOpenVinoTask()
        correct_values_dict = {"dataset": DatasetEntity()}
        unexpected_str = "unexpected string"
        unexpected_values = [
            # Unexpected string is specified as "dataset" parameter
            ("dataset", unexpected_str),
            # Unexpected string is specified as "inference_parameters" parameter
            ("inference_parameters", unexpected_str),
        ]
        check_value_error_exception_raised(
            correct_parameters=correct_values_dict,
            unexpected_values=unexpected_values,
            class_or_function=task.infer,
        )

    @e2e_pytest_unit
    def test_openvino_task_evaluate_params_validation(self):
        """
        <b>Description:</b>
        Check ClassificationOpenVINOTask object "evaluate" method input parameters validation

        <b>Input data:</b>
        ClassificationOpenVINOTask object. "evaluate" method unexpected-type input parameters

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "evaluate" method
        """
        result_set = ResultSetEntity(
            model=model(),
            ground_truth_dataset=DatasetEntity(),
            prediction_dataset=DatasetEntity(),
        )
        task = MockOpenVinoTask()
        correct_values_dict = {"output_result_set": result_set}
        unexpected_int = 1
        unexpected_values = [
            # Unexpected integer is specified as "output_result_set" parameter
            ("output_result_set", unexpected_int),
            # Unexpected integer is specified as "evaluation_metric" parameter
            ("evaluation_metric", unexpected_int),
        ]
        check_value_error_exception_raised(
            correct_parameters=correct_values_dict,
            unexpected_values=unexpected_values,
            class_or_function=task.evaluate,
        )

    @e2e_pytest_unit
    def test_openvino_task_deploy_params_validation(self):
        """
        <b>Description:</b>
        Check ClassificationOpenVINOTask object "deploy" method input parameters validation

        <b>Input data:</b>
        ClassificationOpenVINOTask object. "output_model" non-ModelEntity object

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "deploy" method
        """
        task = MockOpenVinoTask()
        with pytest.raises(ValueError):
            task.deploy(output_model="unexpected string")  # type: ignore

    @e2e_pytest_unit
    def test_openvino_task_optimize_params_validation(self):
        """
        <b>Description:</b>
        Check ClassificationOpenVINOTask object "optimize" method input parameters validation

        <b>Input data:</b>
        ClassificationOpenVINOTask object. "optimize" method unexpected-type input parameters

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "optimize" method
        """
        task = MockOpenVinoTask()
        correct_values_dict = {
            "optimization_type": OptimizationType.POT,
            "dataset": DatasetEntity(),
            "output_model": model(),
        }
        unexpected_str = "unexpected string"
        unexpected_values = [
            # Unexpected string is specified as "optimization_type" parameter
            ("optimization_type", unexpected_str),
            # Unexpected string is specified as "dataset" parameter
            ("dataset", unexpected_str),
            # Unexpected string is specified as "output_model" parameter
            ("output_model", unexpected_str),
            # Unexpected string is specified as "optimization_parameters" parameter
            ("optimization_parameters", unexpected_str),
        ]
        check_value_error_exception_raised(
            correct_parameters=correct_values_dict,
            unexpected_values=unexpected_values,
            class_or_function=task.optimize,
        )


class TestOTXOpenVinoDataLoaderInputParamsValidation:
    @e2e_pytest_unit
    def test_openvino_data_loader_init_params_validation(self):
        """
        <b>Description:</b>
        Check OTXOpenVinoDataLoader object initialization parameters validation

        <b>Input data:</b>
        OTXOpenVinoDataLoader object initialization parameters with unexpected type

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        OTXOpenVinoDataLoader object initialization parameter
        """
        classification_inferencer = MockOpenVinoInferencer()
        correct_values_dict = {
            "dataset": DatasetEntity(),
            "inferencer": classification_inferencer,
        }
        unexpected_str = "unexpected string"
        unexpected_values = [
            # Unexpected string is specified as "dataset" parameter
            ("dataset", unexpected_str),
            # Unexpected string is specified as "inferencer" parameter
            ("inferencer", unexpected_str),
        ]

        check_value_error_exception_raised(
            correct_parameters=correct_values_dict,
            unexpected_values=unexpected_values,
            class_or_function=OTXOpenVinoDataLoader,
        )

    @e2e_pytest_unit
    def test_openvino_data_loader_getitem_input_params_validation(self):
        """
        <b>Description:</b>
        Check OTXOpenVinoDataLoader object "__getitem__" method input parameters validation

        <b>Input data:</b>
        OTXOpenVinoDataLoader object. "__getitem__" method unexpected-type input parameters

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "__getitem__" method
        """
        classification_inferencer = MockOpenVinoInferencer()
        data_loader = OTXOpenVinoDataLoader(dataset=DatasetEntity(), inferencer=classification_inferencer)
        with pytest.raises(ValueError):
            data_loader.__getitem__(index="unexpected string")  # type: ignore


class TestClassificationOpenVINOInferencerInputParamsValidation:
    @e2e_pytest_unit
    def test_openvino_classification_inferencer_init_params_validation(self):
        """
        <b>Description:</b>
        Check ClassificationOpenVINOInferencer object initialization parameters validation

        <b>Input data:</b>
        ClassificationOpenVINOInferencer object initialization parameters

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        ClassificationOpenVINOInferencer object initialization parameter
        """
        correct_values_dict = {
            "hparams": ClassificationConfig("header"),
            "label_schema": LabelSchemaEntity(),
            "model_file": "some model data",
        }
        unexpected_float = 1.1
        unexpected_values = [
            # Unexpected float is specified as "hparams" parameter
            ("hparams", unexpected_float),
            # Unexpected float is specified as "label_schema" parameter
            ("label_schema", unexpected_float),
            # Unexpected float is specified as "model_file" parameter
            ("model_file", unexpected_float),
            # Unexpected float is specified as "weight_file" parameter
            ("weight_file", unexpected_float),
            # Unexpected float is specified as "device" parameter
            ("device", unexpected_float),
            # Unexpected float is specified as "num_requests" parameter
            ("num_requests", unexpected_float),
        ]
        check_value_error_exception_raised(
            correct_parameters=correct_values_dict,
            unexpected_values=unexpected_values,
            class_or_function=ClassificationOpenVINOInferencer,
        )

    @e2e_pytest_unit
    def test_openvino_classification_inferencer_pre_process_params_validation(self):
        """
        <b>Description:</b>
        Check ClassificationOpenVINOInferencer object "pre_process" method input parameters
        validation

        <b>Input data:</b>
        ClassificationOpenVINOInferencer object, "image" non-ndarray object

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "pre_process" method
        """
        inferencer = MockOpenVinoInferencer()
        with pytest.raises(ValueError):
            inferencer.pre_process(image="unexpected string")  # type: ignore

    @e2e_pytest_unit
    def test_openvino_classification_inferencer_post_process_params_validation(self):
        """
        <b>Description:</b>
        Check ClassificationOpenVINOInferencer object "post_process" method input parameters
        validation

        <b>Input data:</b>
        ClassificationOpenVINOInferencer object, "post_process" method unexpected-type input
        parameters

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "post_process" method
        """
        inferencer = MockOpenVinoInferencer()
        correct_values_dict = {
            "prediction": {"prediction_1": np.random.rand(2, 2)},
            "metadata": {"metadata_1": "some_data"},
        }
        unexpected_int = 1
        unexpected_values = [
            # Unexpected integer is specified as "prediction" parameter
            ("prediction", unexpected_int),
            # Unexpected integer is specified as "prediction" dictionary key
            ("prediction", {unexpected_int: np.random.rand(2, 2)}),
            # Unexpected integer is specified as "prediction" dictionary value
            ("prediction", {"prediction_1": unexpected_int}),
            # Unexpected integer is specified as "metadata" parameter
            ("metadata", unexpected_int),
            # Unexpected integer is specified as "metadata" dictionary key
            ("metadata", {unexpected_int: "some_data"}),
        ]
        check_value_error_exception_raised(
            correct_parameters=correct_values_dict,
            unexpected_values=unexpected_values,
            class_or_function=inferencer.post_process,
        )

    @e2e_pytest_unit
    def test_openvino_classification_inferencer_predict_params_validation(self):
        """
        <b>Description:</b>
        Check ClassificationOpenVINOInferencer object "predict" method input parameters
        validation

        <b>Input data:</b>
        ClassificationOpenVINOInferencer object, "image" non-ndarray object

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "predict" method
        """
        inferencer = MockOpenVinoInferencer()
        with pytest.raises(ValueError):
            inferencer.predict(image="unexpected string")  # type: ignore

    @e2e_pytest_unit
    def test_openvino_classification_inferencer_forward_params_validation(self):
        """
        <b>Description:</b>
        Check ClassificationOpenVINOInferencer object "forward" method input parameters validation

        <b>Input data:</b>
        ClassificationOpenVINOInferencer object, "forward" method unexpected-type input parameters

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "forward" method
        """
        inferencer = MockOpenVinoInferencer()
        unexpected_int = 1
        for unexpected_value in [
            # Unexpected integer is specified as "inputs" parameter
            unexpected_int,
            # Unexpected integer is specified as "inputs" dictionary key
            {unexpected_int: np.random.rand(2, 2)},
            # Unexpected integer is specified as "inputs" dictionary value
            {"input_1": unexpected_int},
        ]:
            with pytest.raises(ValueError):
                inferencer.forward(inputs=unexpected_value)  # type: ignore



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from otx.mpa.utils.logger import get_logger

logger = get_logger()


class CDLIterator:
    def __init__(self, cdl):
        self._cdl = cdl
        self._index = 0
        self._cdl_iter = [iter(dl) for dl in self._cdl.loaders]

    def __next__(self):
        if self._index < self._cdl.max_iter:
            batches = {}
            for i, it in enumerate(self._cdl_iter):
                if i == 0:
                    batches = next(it)
                else:
                    try:
                        batches[f"extra_{i-1}"] = next(it)
                    except StopIteration:
                        self._cdl_iter[1] = iter(self._cdl.loaders[1])
                        batches[f"extra_{i-1}"] = next(self._cdl_iter[1])
            self._index += 1
            return batches
        raise StopIteration


class ComposedDL(object):
    class DummySampler(object):
        """dummy sampler class to relay set_epoch() call to the
        list of data loaders in the CDL
        """

        def __init__(self, cdl):
            self.cdl = cdl

        def set_epoch(self, epoch):
            loaders = self.cdl.loaders
            for loader in loaders:
                loader.sampler.set_epoch(epoch)

    def __init__(self, loaders=[]):
        self.loaders = loaders
        self.max_iter = len(self.loaders[0])
        logger.info(f"possible max iterations = {self.max_iter}")
        self._sampler = ComposedDL.DummySampler(self)

    def __len__(self):
        return self.max_iter

    def __iter__(self):
        return CDLIterator(self)

    @property
    def sampler(self):
        return self._sampler



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import importlib

import onnx
from mmcv.utils import ConfigDict


def is_mmdeploy_enabled():
    return importlib.util.find_spec("mmdeploy") is not None


def mmdeploy_init_model_helper(ctx, model_checkpoint=None, cfg_options=None, **kwargs):
    model_builder = kwargs.pop("model_builder")
    model = model_builder(
        ctx.model_cfg,
        checkpoint=model_checkpoint,
        device=ctx.device,
        cfg_options=cfg_options,
    )

    # TODO: Need to investigate it why
    # NNCF compressed model lost trace context from time to time with no reason
    # even with 'torch.no_grad()'. Explicitly setting 'requires_grad' to'False'
    # makes things easier.
    for i in model.parameters():
        i.requires_grad = False

    return model


def update_deploy_cfg(onnx_path, deploy_cfg, mo_options={}):
    from mmdeploy.utils import get_backend_config, get_ir_config

    onnx_model = onnx.load(onnx_path)
    ir_config = get_ir_config(deploy_cfg)
    backend_config = get_backend_config(deploy_cfg)

    # update input
    input_names = [i.name for i in onnx_model.graph.input]
    ir_config["input_names"] = input_names

    # update output
    output_names = [i.name for i in onnx_model.graph.output]
    ir_config["output_names"] = output_names

    # update mo options
    deploy_cfg.merge_from_dict({"backend_config": {"mo_options": mo_options}})


if is_mmdeploy_enabled():
    # fmt: off
    # FIXME: openvino pot library adds stream handlers to root logger
    # which makes annoying duplicated logging
    from mmdeploy.utils import get_root_logger
    get_root_logger().propagate = False
    # fmt: on
